"""
    Copyright (C) 2007-2019, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from PyPi
    https://pypi.org/project/jampy

    If you have found this software useful for your research,
    I would appreciate an acknowledgement and a link to the website.

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

############################################################################


CHANGELOG
---------

    V1.0.0: Written and tested against the corresponding MATLAB version,
        Michele Cappellari, Oxford, 22 October, 2007

    V1.1.0: Allow function parameters to be passed via the FUNCTARGS keyword.
        MC, Oxford, 23 October 2007

    V1.1.1: Added STATUS keyword. Provide more informative error messages on failure.
        MC, Windoek, 5 October 2008

    V1.1.2: Renamed CAP_QUADVA to avoid potential naming conflicts.
        MC, Paranal, 8 November 2013

    V2.0.0: Translated from IDL into Python. MC, Paranal, 14 November 2013

    V2.0.1: Fixed possible program stop. MC, Oxford, 27 January 2014

    V2.0.2: Support both legacy Python 2.7 and Python 3. MC, Oxford, 25 May 2014

    V2.0.3: Dropped Python 2.7 support. MC, Oxford, 12 May 2018

    V2.1.0: Included optional plotting of sampled points. MC, Oxford, 25 June 2019

"""
import numpy as np
import matplotlib.pyplot as plt

EPS = 100*np.MachAr().eps

#----------------------------------------------------------------------
def _qva_split(interval):
    """
    If breakpoints are specified, split subintervals in
    half as needed to get a minimum of 10 subintervals.
    """
    v = interval
    while True:
        npts = interval.size
        if npts > 10: 
            break
        v = np.zeros(npts*2 - 1)
        v[0::2] = interval
        v[1::2] = interval[:-1] + 0.5*np.diff(interval)
        interval = v
    
    return v
#----------------------------------------------------------------------
def _qva_check_spacing(x):

    ax = np.abs(x) 
    too_close = np.any(np.diff(x) <= EPS*np.maximum(ax[:-1], ax[1:]))
    
    return too_close
#----------------------------------------------------------------------
def _qva_f1(fun, t, a, b, args):
    """
    Transform to weaken singularities at both ends: [a, b] -> [-1, 1]
    """
    Tt = 0.25*(b - a)*t*(3. - t**2) + 0.5*(b + a)
    too_close = _qva_check_spacing(Tt)
    if  too_close:
        y = []
    else:
        y = fun(Tt, *args)
        y *= 0.75*(b - a)*(1. - t**2)
        
    return y, too_close
#----------------------------------------------------------------------
def _qva_f2(fun, t, a, b, args):
    """
    Transform to weaken singularity at left end: [a,np.Inf) -> [0,Inf).
    Then transform to finite interval: [0,Inf) -> [0,1].
    """
    Tt = t/(1. - t)
    T2t = a + Tt**2
    too_close = _qva_check_spacing(T2t)
    if  too_close:
        y = []
    else:
        y = fun(T2t, *args)
        y *= 2.*Tt/(1. - t)**2

    return y, too_close
#----------------------------------------------------------------------    
def _qva_f3(fun, t, a, b, args):
    """
    Transform to weaken singularity at right end: (-Inf, b] -> (-Inf, b].
    Then transform to finite interval: (-Inf, b] -> (-1, 0].
    """
    Tt = t/(1. + t)
    T2t = b - Tt**2
    too_close = _qva_check_spacing(T2t)
    if too_close:
        y = []
    else:
        y = fun(T2t, *args)
        y *= -2.*Tt/(1. + t)**2

    return y, too_close
#----------------------------------------------------------------------    
def _qva_f4(fun, t, a, b, args):
    """
    Transform to finite interval: (-Inf,Inf) -> (-1,1)
    """
    Tt = t/(1. - t**2)
    too_close = _qva_check_spacing(Tt)
    if too_close:
        y = []
    else:
        y = fun(Tt, *args)
        y *= (1. + t**2)/(1. - t**2)**2
        
    return y, too_close
#----------------------------------------------------------------------        

def _qva_Vadapt(f, tinterval, rtol, atol, samples, nodes, wt, ewt, fun, a, b, args, plot):

    tbma = np.abs(tinterval[-1] - tinterval[0]) # length of transformed interval
    
    # Initialize array of subintervals of [a,b].
    subs = np.column_stack([tinterval[:-1], tinterval[1:]]) # Two columns array[n, 2]
    
    # Initialize partial sums.
    IfxOK = 0
    errOK = 0
    
    # Initialize main loop
    OK = True
    first = True
    converged = False
    Ifx = 0.
    errbnd = 0.

    if plot:
        plt.figure('quadva')
        plt.clf()
        plt.xlabel('x')
        plt.ylabel('y')
        xall = yall = []

    while True:
        # SUBS contains subintervals of [a,b] where the integral is not
        # sufficiently accurate.  The first row of SUBS holds the left end
        # points and the second row, the corresponding right end points.
        midpt = np.sum(subs, axis=1)/2.      # midpoints of the subintervals
        halfh = np.diff(subs, axis=1)/2.     # half the lengths of the subintervals
        x = nodes * halfh + midpt[:, np.newaxis]  # broadcasting midpt
        halfh = halfh.ravel()
        x = x.ravel()
    
        fx, too_close = f(fun, x, a, b, args)

        if plot:
            plt.figure('quadva')
            plt.plot(x, fx, '+')
            xall = np.append(xall, x)
            yall = np.append(yall, fx)

        # Quit if mesh points are too close or too close to a
        # singular point or got into trouble on first evaluation.
        not_finite = np.any(np.isinf(fx))
        if too_close or not_finite: 
            break
    
        fx = fx.reshape(-1, samples)
    
        # Quantities for subintervals.
        Ifxsubs = fx @ wt * halfh
        errsubs = fx @ ewt * halfh
    
        # Quantities for all of [a, b].
        Ifx = np.sum(Ifxsubs) + IfxOK
        errbnd = abs(np.sum(errsubs) + errOK)
    
        # Test for convergence:
        tol = max(atol, rtol*np.abs(Ifx))
        if errbnd <= tol: 
            converged = True
            break
    
        # Locate subintervals where the approximate integrals are
        # sufficiently accurate and use them to update partial sums.
        good = np.abs(errsubs) <= (2./tbma)*halfh*tol
        errOK += np.sum(errsubs[good])
        IfxOK += np.sum(Ifxsubs[good])
        # Only keep subintervals which are not yet sufficiently accurate
        subs = subs[~good, :]
        if subs.size == 0:   # all intervals are accurate
            converged = True
            break
    
        # Split the remaining subintervals in half. Quit if splitting
        # results in too many subintervals.
        many_subint = 2.*subs.size > 650  # multiplied limit by 10x MC 26/FEB/2008
        if many_subint: 
            break 
        midpt = np.sum(subs, axis=1)/2.
        tmp = np.column_stack([subs[:, 0], midpt, midpt, subs[:, 1]])
        subs = tmp.reshape(-1, 2)  # ---> subs[n, 2]
        first = False

    if first and not converged:
        if too_close: 
            print('***Sub intervals too close.')
        elif not_finite:
            print('***Infinite values in integrand.')
        elif many_subint: 
            print('***Too many sub intervals.')
        OK = False

    if plot:
        plt.figure('quadva')
        w = np.argsort(xall)
        plt.plot(xall[w], yall[w])

    return Ifx, errbnd, OK
#----------------------------------------------------------------------

def quadva(fun, interval, epsrel=1e-5, epsabs=1e-10, plot=False, args=()):
    """
    Calling sequence: ``Ifx, errbnd = quadva(fun,interval,epsrel=1e-5, epsabs=1e-10)``

    When ``INTERVAL = [A,B]``, computes the integral of a continuous function
    ``f(x)`` from A to B for A < B.  A can be ``-np.inf`` and/or B can be
    ``np.inf``.

    FUN accepts a row vector X and returns a vector Y with ``Y(m) = f(X(m))``
    for ``m = 1,...,len(X)``.

    QUADVA returns an approximation Ifx to the integral and optionally an
    approximate bound, ERRBND, on the error ``|integral - Ifx|``. It attempts
    to compute Ifx such that ``|Ifx - integral| <= max(ABSTOL,RELTOL*|Ifx|)``.
    If QUADVA is unsuccessful, Ifx and ERRBND are still meaningful, so a
    warning is issued that includes ERRBND.

    If the interval is infinite, say ``[a, np.inf)``, then for the integral to
    exist, ``f(x)`` must decay as ``x -> np.inf`` and QUADVA requires it to
    decay rapidly. Special methods should be used for oscillatory functions on
    infinite intervals, but QUADVA can be used if ``f(x)`` decays fast enough.

    QUADVA will integrate ``f(x)`` that are singular at finite end points if
    the singularities are not too strong. For example, it will integrate
    ``f(x)`` that behave like ``log|x - c|`` or ``|x - c|^p`` for ``p >= -1/2``
    with ``c = a`` and/or ``c = b``. If ``f(x)`` is singular at points inside
    ``(A,B)``, write the integral as a sum of integrals over subintervals with
    the singular points as end points, compute them with QUADVA, and add the
    results.

    QUADVA starts with samples of ``f(x)`` at 150 points in ``(A,B)``. It must
    be able to recognize the behavior of ``f(x)`` from these samples, so if
    ``f(x)`` oscillates very rapidly or has sharp peaks, it may be necessary to
    subdivide the interval. To do this, make INTERVAL an array with entries
    (breakpoints) that increase from ``A = INTERVAL[0]`` to ``B = INTERVAL[-1]``.
    If ``f(x)`` is only piecewise smooth, the points of discontinuity should
    be breakpoints.

    Based on the algorithm "Vectorized Adaptive Quadrature in Matlab"
    `L.F. Shampine, Journal of Computational and Applied Mathematics
    211 (2008) 131-140 <http://dx.doi.org/10.1016/j.cam.2006.11.021>`_

    """

    interval = np.asarray(interval)    
    nint = interval.size
    assert nint >= 2, "INTERVAL must be a real vector of at least two entries."
    assert np.all(np.diff(interval) > 0), "Entries of INTERVAL must strictly increase."

    a = interval[0]
    b = interval[-1]
    
    # Generally the error test is a mixed one, but pure absolute error
    # and pure relative error are allowed.  If a pure relative error
    # test is specified, the tolerance must be at least 100*EPS. 
    #
    rtol = max(epsrel, EPS)
    atol = epsabs

    # Gauss-Kronrod (7,15) pair. Use symmetry in defining nodes and weights.
    #
    samples = 15
    pnodes = np.array([0.2077849550078985, 0.4058451513773972, 0.5860872354676911,
                       0.7415311855993944, 0.8648644233597691, 0.9491079123427585, 
                       0.9914553711208126])
    nodes = np.hstack([-pnodes[::-1], 0, pnodes])
    pwt = np.array([0.2044329400752989, 0.1903505780647854, 0.1690047266392679, 
                    0.1406532597155259, 0.1047900103222502, 0.06309209262997855, 
                    0.02293532201052922])
    wt = np.hstack([pwt[::-1], 0.2094821410847278, pwt])
    pwt7 = np.array([0, 0.3818300505051189, 0, 0.2797053914892767, 
                     0, 0.1294849661688697, 0])
    ewt = wt - np.hstack([pwt7[::-1], 0.4179591836734694, pwt7])
    
    # Identify the task. If breakpoints are specified, work out
    # how they map into the standard interval.
    #
    if a != -np.Inf and b != np.Inf:
        if nint > 2:
            # Analytical transformation suggested by K.L. Metlov:
            alpha = 2.*np.sin(np.arcsin((a + b - 2.*interval[1:-1])/(a - b))/3.)
            tinterval = np.hstack([-1., alpha, 1.])
            tinterval = _qva_split(tinterval)
        else:
            tinterval = np.linspace(-1, 1, 11)
        Ifx,errbnd,OK = _qva_Vadapt(_qva_f1, tinterval, rtol, atol, samples,
                                    nodes, wt, ewt, fun, a, b, args, plot)
    elif a != -np.Inf and b == np.Inf:
        if nint > 2:
            alpha = np.sqrt(interval[1:-1] - a)
            tinterval = np.hstack([0., alpha/(1. + alpha), 1.])
            tinterval = _qva_split(tinterval)
        else:
            tinterval = np.linspace(0, 1, 11)
        Ifx,errbnd,OK = _qva_Vadapt(_qva_f2,tinterval,rtol,atol,samples,
                                   nodes,wt,ewt,fun,a,b,args, plot)
    elif a == -np.Inf and b != np.Inf:
        if nint > 2:
            alpha = np.sqrt(b - interval[1:-1])
            tinterval = np.hstack([-1., -alpha/(1. + alpha), 0.])
            tinterval = _qva_split(tinterval)
        else:
            tinterval = np.linspace(-1, 0, 11)
        Ifx,errbnd,OK = _qva_Vadapt(_qva_f3, tinterval, rtol, atol, samples,
                                    nodes, wt, ewt, fun, a, b, args, plot)
    elif a == -np.Inf and b == np.Inf:
        if nint > 2:
            # Analytical transformation suggested by K.L. Metlov:
            alpha = np.tanh(np.arcsinh(2.*interval[1:-1])/2.)
            tinterval = np.hstack([-1., alpha, 1.])
            tinterval = _qva_split(tinterval)
        else:
            tinterval = np.linspace(-1, 1, 11)
        Ifx,errbnd,OK = _qva_Vadapt(_qva_f4, tinterval, rtol, atol, samples,
                                    nodes,wt,ewt,fun,a,b,args, plot)
    
    if OK:
        status = 0 # success   
    else: 
        print('***Integral does not satisfy error test.')
        print('***Approximate bound on error is', errbnd)
        status = 1
    
    return Ifx, errbnd, status
#----------------------------------------------------------------------        
def quadva_test1(x, a=3, b=5):
    # Gladwell's problem no1. limits x=[0,8]
    # Precise result: 0.33333333332074955152
    
    return np.exp(-a*x)-np.cos(b*np.pi*x)
#----------------------------------------------------------------------        
def quadva_test2(x):
    # Gladwell's problem no2. limits x=[-1,2]
    # Precise result: 5.9630898453302550932
    
    return np.abs(x - 1/np.sqrt(3)) + np.abs(x + 1/np.sqrt(2))
#----------------------------------------------------------------------
def quadva_test3(x, a=2.0/3):
    # Gladwell's problem no3. limits x=[0,1]
    # Precise result: 3 with x^(-2d/3d)
    
    return x**a
#----------------------------------------------------------------------
def quadva_test4(x):
    # Gladwell's problem no3. limits x=[0,1]
    # Precise result: 2.5066282746310005024
    
    return np.exp(-x**2/2)
#----------------------------------------------------------------------
def quadva_examples():

    print('\ntest 1 ###############################\n')
    ifx, errbnd, ok = quadva(quadva_test1, [0, 8], args=(3, 5), epsrel=0, epsabs=1e-12)
    print(ifx, '+/-', errbnd, ok) 
    print('actual error:', ifx - 0.33333333332074955152)
    print('\ntest 2 ###############################\n')
    ifx, errbnd, ok = quadva(quadva_test2, [-1, 2], epsrel=0, epsabs=1e-12)
    print(ifx, '+/-', errbnd, ok) 
    print('actual error:', ifx - 5.9630898453302550932)
    print('\ntest 2 with breakpoints ##############\n')
    ifx, errbnd, ok = quadva(quadva_test2, [-1, -1/np.sqrt(2), 1/np.sqrt(3), 2], epsrel=0, epsabs=1e-12)
    print(ifx, '+/-', errbnd, ok) 
    print('actual error:', ifx - 5.9630898453302550932)
    print('\ntest 3 ###############################\n')
    ifx, errbnd, ok = quadva(quadva_test3, [0, 1], args=(-2.0/3,), epsrel=0, epsabs=1e-12)
    print(ifx, '+/-', errbnd, ok) 
    print('actual error:', ifx - 3)
    print('\ntest 4 ###############################\n')
    ifx, errbnd, ok = quadva(quadva_test4, [-np.Inf, np.Inf], epsrel=1e-5, epsabs=0)
    print(ifx, '+/-', errbnd, ok) 
    print('actual error:', ifx - 2.5066282746310005024)
    ifx, errbnd, ok = quadva(quadva_test4, [-6, 6], epsrel=1e-5, epsabs=0, plot=1)
    print(ifx, '+/-', errbnd, ok) 
    print('actual error:', ifx - 2.5066282746310005024)
#----------------------------------------------------------------------

if __name__ == '__main__':
    quadva_examples()



"""
    Copyright (C) 2019-2020, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from my web page
    http://purl.org/cappellari/software

CHANGELOG
---------

V1.1.0: MC, Oxford, 16 July 2020
    - Compute both Vrms and LOS velocity.
V1.0.1: MC, Oxford, 21 April 2020
    - Made a separate file
V1.0.0: Michele Cappellari, Oxford, 08 November 2019
    - Written and tested

"""
import numpy as np
import matplotlib.pyplot as plt

from jampy.jam_axi_proj import jam_axi_proj

##############################################################################

def jam_axi_proj_example():
    """
    Usage example for jam_axi_proj.
    It takes about 2s on a 3GHz CPU

    """
    np.random.seed(123)
    xbin, ybin = np.random.uniform(low=[-55, -40], high=[55, 40], size=[1000, 2]).T

    inc = 60.                                                # Assumed galaxy inclination
    r = np.sqrt(xbin**2 + (ybin/np.cos(np.radians(inc)))**2) # Radius in the plane of the disk
    a = 40                                                   # Scale length in arcsec
    vr = 2000*np.sqrt(r)/(r+a)                               # Assumed velocity profile (v_c of Hernquist 1990)
    vel = vr * np.sin(np.radians(inc))*xbin/r                # Projected velocity field
    sig = 8700/(r+a)                                         # Assumed velocity dispersion profile
    rms = np.sqrt(vel**2 + sig**2)                           # Vrms field in km/s

    # Until here I computed some fake input kinematics to fit with JAM.
    # Ina real application, instead of the above lines one will read the
    # measured stellar kinematics, e.g. from integral-field spectroscopy

    surf = np.array([39483., 37158., 30646., 17759., 5955.1, 1203.5, 174.36, 21.105, 2.3599, 0.25493])
    sigma = np.array([0.153, 0.515, 1.58, 4.22, 10, 22.4, 48.8, 105, 227, 525])
    qObs = np.full_like(sigma, 0.57)

    distance = 16.5     # Assume Virgo distance in Mpc (Mei et al. 2007)
    mbh = 1e8           # Black hole mass in solar masses
    beta = np.full_like(surf, 0.2)

    # Below I assume mass follows light, but in a real application one
    # will generally include a dark halo in surf_pot, sigma_pot, qobs_pot.
    # See e.g. Cappellari (2013) for an example
    # https://ui.adsabs.harvard.edu/abs/2013MNRAS.432.1709C

    surf_lum = surf_pot = surf
    sigma_lum = sigma_pot = sigma
    qobs_lum = qobs_pot = qObs

    sigmapsf = [0.6, 1.2]
    normpsf = [0.7, 0.3]
    pixsize = 0.8
    goodbins = r > 10  # Arbitrarily exclude the center to illustrate how to use goodbins

    # I use a loop below, just to higlight the fact that all parameters
    # remain the same for the two JAM calls, except for 'moment' and 'data'
    for moment, data in zip(['zz', 'z'], [rms, vel]):

        # The model is by design similar but not identical to the adopted kinematics!
        m = jam_axi_proj(surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
                         inc, mbh, distance, xbin, ybin, plot=True, data=data,
                         sigmapsf=sigmapsf, normpsf=normpsf, beta=beta, pixsize=pixsize,
                         moment=moment, goodbins=goodbins, align='cyl', ml=None)
        plt.pause(1)
        surf_pot *= m.ml  # Scale the density by the best fitting M/L from the previous step

##############################################################################

if __name__ == '__main__':

    jam_axi_proj_example()



#!/usr/bin/env python
"""
    In this example/test we fit a one dimensional MGE to the density
    of a Hernquist (1990, ApJ, 356, 359; hereafter H90) model.
    - IMPORTANT: One needs the MGE_FIT_1D routine from the `mgefit`
    package available from http://purl.org/cappellari/software.
    We then compute the following quantities for a spherical model:
    (i) The circular velocity with MGE_VCIRC
    (ii) The sigma of an isotropic model, with both the JAM_SPH_RMS
        and JAM_AXI_RMS routines;
    (iii) The sigma of a fully tangential anisotropic model with both the
        JAM_SPH_RMS and JAM_AXI_RMS routines.
    The solutions are compared with the analytic results by H90 and
    with the solution presented in Fig.4.20 of Binney & Tremaine (2008).

    V1.0.0: Michele Cappellari, Oxford, 28 November 2008
    V1.1.0: Included test with anisotropy and black holes.
      MC, Oxford, 17 June 2010
    V2.0.0: Translated from IDL into Python. MC, Oxford, 9 April 2014
    V2.0.1: Fixed RuntimeWarning. MC, Oxford, 17 March 2017
    V2.0.2: Changed imports for jam as a package. MC, Oxford, 17 April 2018    

"""

import numpy as np
import matplotlib.pyplot as plt

from mgefit.mge_fit_1d import mge_fit_1d
from jampy.mge_vcirc import mge_vcirc
from jampy.legacy.jam_axi_rms import jam_axi_rms
from jampy.legacy.jam_sph_rms import jam_sph_rms

#----------------------------------------------------------------------------

def jam_hernquist_model_example():

    M = 1e11          # Total mass in Solar Masses
    a = 1e3           # Break radius in pc
    distance = 16.5   # Assume Virgo distance in Mpc (Mei et al. 2007)
    pc = distance*np.pi/0.648 # Constant factor to convert arcsec --> pc
    G = 0.004301      # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA-14)]
    mbh = 0

    n = 300  # Number of values to sample the H90 profile for the fit
    r = np.logspace(-2.5, 2, n)*a   # logarithmically spaced radii in pc
    rho = M*a/(2*np.pi*r*(r+a)**3)  # Density in Msun/pc**3 (H90 equation 2)

    m = mge_fit_1d(r, rho, ngauss=16, plot=True)
    plt.pause(1)

    surf = m.sol[0]       # Surface density in Msun/pc**2
    sigma = m.sol[1]/pc   # Gaussian dispersion in arcsec
    qObs = np.full_like(surf, 1)         # Assume spherical model
    inc = 90                # Edge-on view
    npix = 100
    rad = np.linspace(0.01, 50, npix) # desired output radii in arcsec (avoid R=0)
    r = rad*pc

    ################## Circular Velocity #################

    vcirc = mge_vcirc(surf, sigma, qObs, inc, mbh, distance, rad)

    plt.clf()
    plt.subplot(211)
    plt.plot(rad, vcirc, 'b', linewidth=8)
    plt.xlabel('R (arcsec)')
    plt.ylabel(r'$V_{circ}$, $V_{rms}$ (km/s)')
    plt.title('Reproduces the analytic results by Hernquist (1990)')

    # Compare with analytic result
    #
    vc = np.sqrt(G*M*r)/(r+a) # H90 equation (16)
    plt.plot(rad, vc, 'LimeGreen', linewidth=3)
    plt.text(30, 310, '$V_{circ}$ (mge_vcirc, H90)')

    #################### Isotropic Vrms ###################

    # Spherical isotropic H90 model
    #
    sigp = jam_sph_rms(surf, sigma, surf, sigma, mbh, distance, rad)[0]
    plt.plot(rad, sigp, 'b', linewidth=8)

    # Axisymmetric isotropic model in the spherical limit.
    # This is plotted on the major axis, but the Vrms has circular symmetry
    #
    vrms = jam_axi_rms(surf, sigma, qObs, surf, sigma, qObs,
                       inc, mbh, distance, rad, rad*0)[0]
    plt.plot(rad, vrms, 'r', linewidth=5)

    # Analytic surface brightness from H90
    #
    s = r/a
    w = s < 1
    xs = np.hstack([np.arccosh(1/s[w]) / np.sqrt(1 - s[w]**2),    # H90 eq. (33)
                    np.arccos(1/s[~w]) / np.sqrt(s[~w]**2 - 1)])  # H90 eq. (34)
    IR = M*((2 + s**2)*xs - 3) / (2*np.pi*a**2*(1 - s**2)**2)     # H90 eq. (32)

    # Projected second moments of isotropic model from H90
    #
    sigp = np.sqrt(G*M**2/(12*np.pi*a**3*IR) # H90 equation (41)
                 *(0.5/(1 - s**2)**3
                 *(-3*s**2*xs*(8*s**6 - 28*s**4 + 35*s**2 - 20)
                 - 24*s**6 + 68*s**4 - 65*s**2 + 6) - 6*np.pi*s))
    plt.plot(rad, sigp, 'LimeGreen', linewidth=2)
    plt.text(20, 90, 'Isotropic $V_{rms}$ (jam_sph, jam_axi, H90)')

    ################### Anisotropic Vrms ##################

    # Projected second moments for a H90 model with sigma_R=0.
    # This implies beta=-Infinity but I adopt as an approximation
    # below a large negative beta. This explains why te curves do not
    # overlap perfectly.
    #
    beta = np.full_like(surf, -20)
    sigp = jam_sph_rms(surf, sigma, surf, sigma, mbh, distance, rad, beta=beta)[0]
    plt.plot(rad, sigp, linewidth=8)

    # Axisymmetric anisotropic model in the spherical limit.
    # The spherical Vrms is the quadratic average of major
    # and minor axes of the axisymmetric model.
    #
    vrms_maj = jam_axi_rms(surf, sigma, qObs, surf, sigma, qObs,
                           inc, mbh, distance, rad, rad*0, beta=beta)[0]
    vrms_min = jam_axi_rms(surf, sigma, qObs, surf, sigma, qObs,
                           inc, mbh, distance, rad*0, rad, beta=beta)[0]
    plt.plot(rad, np.sqrt((vrms_maj**2 + vrms_min**2)/2), 'r', linewidth=5)

    # Projected second moments of fully tangential model from H90
    #
    sigp = np.sqrt(G*M**2*r**2/(2*np.pi*a**5*IR) # H90 equation (42)
                 *(1./(24*(1 - s**2)**4)
                 *(-xs*(24*s**8 - 108*s**6 + 189*s**4 - 120*s**2 + 120)
                 - 24*s**6 + 92*s**4 - 117*s**2 + 154) + 0.5*np.pi/s))
    plt.plot(rad, sigp, 'LimeGreen', linewidth=2)
    plt.text(20, 200, 'Tangential $V_{rms}$ (jam_sph, jam_axi, H90)')

    ############### Anisotropic models with Black Hole ###############

    # Reproduces Fig.4.20 of Binney J., & Tremaine S.D., 2008,
    # Galactic Dynamics, 2nd ed., Princeton University Press
    # See https://books.google.co.uk/books?id=6mF4CKxlbLsC&pg=PA352

    plt.subplot(212)
    plt.xlabel('R/a')
    plt.ylabel('$\\sigma\, (G M / a)^{-1/2}$')
    plt.title('Reproduces Fig.4.20 of Binney & Tremaine (2008)')

    cost = np.sqrt(G*M/a)
    rad = np.logspace(-2.3, 1, 50)*a/pc # desired output radii in arcsec
    bhs = np.array([0., 0.002, 0.004])*M
    betas = np.array([-0.51, 0, 0.51]) # Avoids singularity at beta=+/-0.5
    colors = ['r', 'b', 'LimeGreen']

    for color, beta in zip(colors, betas):
        betaj = np.full_like(surf, beta)
        for bh in bhs:
            sigp = jam_sph_rms(surf, sigma, surf, sigma, bh, distance, rad, beta=betaj)[0]
            plt.semilogx(rad/a*pc, sigp/cost, linewidth=5, color=color)

    # Test the jam_axi_rms with Black Hole in the spherical isotropic limit
    #
    for bh in bhs:
        vrms = jam_axi_rms(surf, sigma, qObs, surf, sigma, qObs,
                           inc, bh, distance, rad, rad*0)[0]
        plt.semilogx(rad/a*pc, vrms/cost, linewidth=2, color=colors[2])

    plt.axis([0.006, 10, 0, 0.6]) # x0, x1, y0, y1
    plt.tight_layout()

    plt.text(0.03, 0.15,'$\\beta$=-0.5 (tangential: jam_sph)', color=colors[0])
    plt.text(.03, 0.35,'$\\beta$=0 (isotropic: jam_sph, jam_axi)', color=colors[1])
    plt.text(0.03, 0.5,'$\\beta$=0.5 (radial: jam_sph)', color=colors[2])
    plt.pause(1)  # allow the plot to appear in certain situations

#----------------------------------------------------------------------------

if __name__ == '__main__':
    jam_hernquist_model_example()



"""
    Copyright (C) 2003-2019, Michele Cappellari
    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from my web page
    http://purl.org/cappellari/software

    If you have found this software useful for your research,
    I would appreciate an acknowledgement to the use of the
    "JAM modelling method of Cappellari (2008)"

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

"""

import numpy as np
import matplotlib.pyplot as plt
from scipy import special, signal, ndimage
from time import perf_counter as clock

from plotbin.plot_velfield import plot_velfield
from plotbin.symmetrize_velfield import symmetrize_velfield
from jampy.quadva import quadva

##############################################################################

def mge_surf(x, y, surf, sigma, qobs):
    """ MGE surface brightness for a set of coordinates (x, y) """

    mge = surf*np.exp(-0.5/sigma**2*(x[..., None]**2 + (y[..., None]/qobs)**2))

    return mge.sum(-1)

##############################################################################

def bilinear_interpolate(xv, yv, im, xout, yout):
    """
    The input array has size im[ny,nx] as in the output
    of im = f(meshgrid(xv, yv))
    xv and yv are vectors of size nx and ny respectively.

    """
    ny, nx = np.shape(im)
    assert (nx, ny) == (xv.size, yv.size), "Input arrays dimensions do not match"

    xi = (nx - 1.)/(xv[-1] - xv[0]) * (xout - xv[0])
    yi = (ny - 1.)/(yv[-1] - yv[0]) * (yout - yv[0])

    return ndimage.map_coordinates(im.T, [xi, yi], order=1, mode='nearest')

##############################################################################

def rotate_points(x, y, ang):
    """
    Rotates points conter-clockwise by an angle ANG in degrees.
    Michele cappellari, Paranal, 10 November 2013

    """
    theta = np.radians(ang)
    xNew = x*np.cos(theta) - y*np.sin(theta)
    yNew = x*np.sin(theta) + y*np.cos(theta)

    return xNew, yNew  

##############################################################################

def integrand(u1,
               dens_lum, sigma_lum, q_lum,
               dens_pot, sigma_pot, q_pot,
               x1, y1, inc, beta, tensor):
    """
    This routine computes the integrand of Eq.(28) of Cappellari (2008; C08) for
    a model with constant anisotropy sigma_R**2 = b*sigma_z**2 and <V_R*V_z> = 0.
    The components of the proper motion dispersions tensor are calculated as
    described in note 5 of C08.
    See Cappellari (2012; C12 http://arxiv.org/abs/1211.7009)
    for explicit formulas for the proper motion tensor.

    """
    dens_lum = dens_lum[:, None, None]
    sigma_lum = sigma_lum[:, None, None]
    q_lum = q_lum[:, None, None]
    beta = beta[:, None, None]

    dens_pot = dens_pot[None, :, None]
    sigma_pot = sigma_pot[None, :, None]
    q_pot = q_pot[None, :, None]

    u = u1[None, None, :]

    kani = 1./(1 - beta) # Anisotropy ratio b = (sig_R/sig_z)**2
    ci = np.cos(inc)
    si = np.sin(inc)
    si2 = si**2
    ci2 = ci**2
    x2 = x1**2
    y2 = y1**2
    u2 = u**2

    s2_lum = sigma_lum**2
    q2_lum = q_lum**2
    e2_lum = 1 - q2_lum
    s2q2_lum = s2_lum*q2_lum

    s2_pot = sigma_pot**2
    e2_pot = 1 - q_pot**2

    # Double summation over (j, k) of eq.(28) for all values of integration variable u.
    # The triple loop in (j, k, u) is replaced by broadcast Numpy array operations.
    e2u2_pot = e2_pot*u2
    a = 0.5*(u2/s2_pot + 1/s2_lum)               # equation (29) in C08
    b = 0.5*(e2u2_pot*u2/(s2_pot*(1 - e2u2_pot)) + e2_lum/s2q2_lum) # equation (30) in C08
    c = e2_pot - s2q2_lum/s2_pot                  # equation (22) in C08
    d = 1 - kani*q2_lum - ((1 - kani)*c + e2_pot*kani)*u2 # equation (23) in C08
    e = a + b*ci2
    if tensor == 'xx':
        f = kani*s2q2_lum + d*((y1*ci*(a+b)/e)**2 + si2/(2*e)) # equation (4) in C12
    elif tensor == 'yy':
        f = s2q2_lum*(si2 + kani*ci2) + d*x2*ci2  # equation (5) in C12
    elif tensor == 'zz':
        f = s2q2_lum*(ci2 + kani*si2) + d*x2*si2  # z' LOS equation (28) in C08
    elif tensor == 'xy':
        f = -d*np.abs(x1*y1)*ci2*(a+b)/e          # equation (6) in C12
    elif tensor == 'xz':
        f = -d*np.abs(x1*y1)*si*ci*(a+b)/e         # -equation (7) in C12
    elif tensor == 'yz':
        f = -si*ci*(s2q2_lum*(1 - kani) - d*x2)   # -equation (8) in C12

    # arr has the dimensions (q_lum.size, q_pot.size, u.size)

    arr = q_pot*dens_pot*dens_lum*u2*f*np.exp(-a*(x2 + y2*(a + b)/e)) \
        / ((1. - c*u2)*np.sqrt((1. - e2u2_pot)*e))

    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA-14)]

    return 4*np.pi**1.5*G*arr.sum((0, 1))

##############################################################################

def surf_v2los(x1, y1, inc,
               dens_lum, sigma_lum, qintr_lum,
               dens_pot, sigma_pot, qintr_pot,
               beta, tensor):
    """
    This routine gives the projected weighted second moment Sigma*<V^2_los>

    """
    sb_mu2 = np.empty_like(x1)
    for j, (xj, yj) in enumerate(zip(x1, y1)):
        sb_mu2[j] = quadva(integrand, [0., 1.],
                            args=(dens_lum, sigma_lum, qintr_lum,
                                  dens_pot, sigma_pot, qintr_pot,
                                  xj, yj, inc, beta, tensor))[0]

    return sb_mu2

##############################################################################

def vrms2(x, y, inc_deg,
           surf_lum, sigma_lum, qobs_lum,
           surf_pot, sigma_pot, qobs_pot,
           beta, tensor, sigmaPsf, normPsf,
           pixSize, pixAng, step, nrad, nang):
    """
    This routine gives the second V moment after convolution with a PSF.
    The convolution is done using interpolation of the model on a
    polar grid, as described in Appendix A of Cappellari (2008).

    """
    # Axisymmetric deprojection of both luminous and total mass.
    # See equation (12)-(14) of Cappellari (2008)
    #
    inc = np.radians(inc_deg)

    qintr_lum = qobs_lum**2 - np.cos(inc)**2
    if np.any(qintr_lum <= 0):
        raise RuntimeError('Inclination too low q <= 0')
    qintr_lum = np.sqrt(qintr_lum)/np.sin(inc)
    if np.any(qintr_lum < 0.05):
        raise RuntimeError('q < 0.05 components')
    dens_lum = surf_lum*qobs_lum / (sigma_lum*qintr_lum*np.sqrt(2*np.pi))

    qintr_pot = qobs_pot**2 - np.cos(inc)**2
    if np.any(qintr_pot <= 0):
        raise RuntimeError('Inclination too low q <= 0')
    qintr_pot = np.sqrt(qintr_pot)/np.sin(inc)
    if np.any(qintr_pot < 0.05):
        raise RuntimeError('q < 0.05 components')
    dens_pot = surf_pot*qobs_pot / (sigma_pot*qintr_pot*np.sqrt(2*np.pi))

    # Define parameters of polar grid for interpolation
    #
    w = sigma_lum < np.max(np.abs(x)) # Characteristic MGE axial ratio in observed range

    if w.sum() < 3:
        qmed = np.median(qobs_lum)
    else:
        qmed = np.median(qobs_lum[w])

    rell = np.sqrt(x**2 + (y/qmed)**2)  # Elliptical radius of input (x, y)

    psfConvolution = (np.max(sigmaPsf) > 0) and (pixSize > 0)

    # Kernel step is 1/4 of largest value between sigma(min) and 1/2 pixel side.
    # Kernel half size is the sum of 3*sigma(max) and 1/2 pixel diagonal.
    if (nrad*nang > x.size) and (not psfConvolution): # Just calculate values

        xPol = x
        yPol = y

    else:  # Interpolate values on polar grid

        if psfConvolution:   # PSF convolution
            if step == 0:
                step = np.min(sigmaPsf)/4.
            mx = 3*np.max(sigmaPsf) + pixSize/np.sqrt(2)
        else:                                   # No convolution
            step = max(np.sqrt(2), np.min(rell))  # Minimum radius of 1pc
            mx = 0

        # Make linear grid in log of elliptical radius RAD and eccentric anomaly ANG
        # See Appendix A
        rmax = np.max(rell) + mx  # Major axis of ellipse containing all data + convolution
        logRad = np.linspace(np.log(step/np.sqrt(2)), np.log(rmax), nrad)  # Linear grid in np.log(rell)
        ang = np.linspace(0, np.pi/2, nang)  # Linear grid in eccentric anomaly
        radGrid, angGrid = map(np.ravel, np.meshgrid(np.exp(logRad), ang))
        xPol = radGrid*np.cos(angGrid)
        yPol = radGrid*np.sin(angGrid) * qmed

    # The model Vrms computation is only performed on the polar grid
    # which is then used to interpolate the values at any other location
    wm2Pol = surf_v2los(xPol, yPol, inc,
                        dens_lum, sigma_lum, qintr_lum,
                        dens_pot, sigma_pot, qintr_pot,
                        beta, tensor)

    if psfConvolution:  # PSF convolution

        nx = int(np.ceil(rmax/step))
        ny = int(np.ceil(rmax*qmed/step))
        x1 = np.linspace(0.5 - nx, nx - 0.5, 2*nx)*step
        y1 = np.linspace(0.5 - ny, ny - 0.5, 2*ny)*step
        xCar, yCar = np.meshgrid(x1, y1)  # Cartesian grid for convolution

        # Interpolate MGE model and Vrms over cartesian grid
        r1 = 0.5*np.log(xCar**2 + (yCar/qmed)**2) # Log elliptical radius of cartesian grid
        e1 = np.arctan2(np.abs(yCar/qmed), np.abs(xCar))    # Eccentric anomaly of cartesian grid

        # # Division by mgePol before interpolation reduces interpolation error
        mgeCar = mge_surf(xCar, yCar, surf_lum, sigma_lum, qobs_lum)
        mgePol = mge_surf(xPol, yPol, surf_lum, sigma_lum, qobs_lum)
        wm2Car = mgeCar*bilinear_interpolate(logRad, ang, (wm2Pol/mgePol).reshape(nang, nrad), r1, e1)

        if tensor in ('xy', 'xz'):
            wm2Car *= np.sign(xCar*yCar)  # Calculation was done in positive quadrant

        nk = int(np.ceil(mx/step))
        kgrid = np.linspace(-nk, nk, 2*nk + 1)*step
        xgrid, ygrid = np.meshgrid(kgrid, kgrid)  # Kernel is square
        if pixAng != 0:
            xgrid, ygrid = rotate_points(xgrid, ygrid, pixAng)

        # Compute kernel with equation (A6) of Cappellari (2008).
        # Normalization is irrelevant here as it cancels out.
        #
        dx = pixSize/2
        sp = np.sqrt(2)*sigmaPsf
        xg, yg = xgrid[..., None], ygrid[..., None]
        kernel = normPsf*(special.erf((dx - xg)/sp) + special.erf((dx + xg)/sp)) \
                        *(special.erf((dx - yg)/sp) + special.erf((dx + yg)/sp))
        kernel = kernel.sum(2)   # Sum over PSF components

        # Seeing and aperture convolution with equation (A3)
        #
        muCar = signal.fftconvolve(wm2Car, kernel, mode='same') \
              / signal.fftconvolve(mgeCar, kernel, mode='same')

        # Interpolate convolved image at observed apertures.
        # Aperture integration was already included in the kernel.
        #
        mu = bilinear_interpolate(x1, y1, muCar, x, y)

    else:  # No PSF convolution

        muPol = wm2Pol/mge_surf(xPol, yPol, surf_lum, sigma_lum, qobs_lum)

        if nrad*nang > x.size:      # Just returns values
            mu = muPol
        else:                      # Interpolate values
            r1 = 0.5*np.log(x**2 + (y/qmed)**2) # Log elliptical radius of input (x,y)
            e1 = np.arctan2(np.abs(y/qmed), np.abs(x))    # Eccentric anomaly of input (x,y)
            mu = bilinear_interpolate(logRad, ang, muPol.reshape(nang, nrad), r1, e1)

            if tensor in ('xy', 'xz'):
                mu *= np.sign(x*y)  # Calculation was done in positive quadrant

    return mu, psfConvolution

##############################################################################

def jam_axi_rms(surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
                inc, mbh, distance, xbin, ybin, beta=None, erms=None,
                flux_obs=None, goodbins=None, ml=None, nang=10, normpsf=1.,
                nrad=20, pixang=0., pixsize=0., plot=True, quiet=False,
                rbh=0.01, rms=None, sigmapsf=0., step=0., tensor='zz',
                vmax=None, vmin=None, **kwargs):

    """

    Purpose of jam_axi_rms
    ----------------------

    This procedure calculates a prediction for the projected second
    velocity moment ``V_RMS = sqrt(V**2 + sigma**2)``, or optionally any
    of the six components of the symmetric proper motion dispersion
    tensor, for an anisotropic (three integrals) axisymmetric galaxy model.
    It implements the solution of the anisotropic Jeans equations presented
    in equation (28) and note 5 of `Cappellari (2008).
    <https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C>`_
    PSF convolution in done as described in the Appendix of that paper.
    See `Cappellari (2012; C12) <http://arxiv.org/abs/1211.7009>`_
    for explicit formulas for the full proper motion tensor.

    Calling Sequence
    ----------------

    .. code-block:: python

        rmsModel, ml, chi2, flux = jam_axi_rms(
                surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
                inc, mbh, distance, xbin, ybin, beta=None, erms=None,
                flux_obs=None, goodbins=None, ml=None, nang=10, normpsf=1.,
                nrad=20, pixang=0., pixsize=0., plot=True, quiet=False,
                rbh=0.01, rms=None, sigmapsf=0., step=0., tensor='zz',
                vmax=None, vmin=None)

    Input Parameters
    ----------------

    SURF_LUM:
        vector of length ``N`` containing the peak surface brightness of the
        MGE Gaussians describing the galaxy surface brightness in units of
        ``Lsun/pc**2`` (solar luminosities per parsec**2).
    SIGMA_LUM:
        vector of length ``N`` containing the dispersion in arcseconds of
        the MGE Gaussians describing the galaxy surface brightness.
    QOBS_LUM:
        vector of length ``N`` containing the observed axial ratio of the MGE
        Gaussians describing the galaxy surface brightness.
    SURF_POT:
        vector of length ``M`` containing the peak value of the MGE Gaussians
        describing the galaxy surface density in units of ``Msun/pc**2`` (solar
        masses per parsec**2). This is the MGE model from which the model
        potential is computed.

        In a common usage scenario, with a self-consistent model, one has
        the same Gaussians for both the surface brightness and the potential.
        This implies ``SURF_POT = SURF_LUM``, ``SIGMA_POT = SIGMA_LUM`` and
        ``QOBS_POT = QOBS_LUM``. The global M/L of the model is fitted by the
        routine when passing the ``RMS`` and ``ERMS`` keywords with the observed
        kinematics.
    SIGMA_POT:
        vector of length ``M`` containing the dispersion in arcseconds of
        the MGE Gaussians describing the galaxy surface density.
    QOBS_POT:
        vector of length ``M`` containing the observed axial ratio of the MGE
        Gaussians describing the galaxy surface density.
    INC:
        inclination in degrees (90 being edge-on).
    MBH:
        Mass of a nuclear supermassive black hole in solar masses.

        VERY IMPORTANT: The model predictions are computed assuming SURF_POT
        gives the total mass. In the common self-consistent case one has
        ``SURF_POT = SURF_LUM`` and if requested (keyword ML) the program can
        scale the output ``RMSMODEL`` to best fit the data. The scaling is
        equivalent to multiplying *both* SURF_POT and MBH by a factor M/L.
        To avoid mistakes, the actual MBH used by the output model is printed
        on the screen.
    DISTANCE:
        distance of the galaxy in Mpc.
    XBIN:
        Vector of length ``P`` with the X coordinates in arcseconds of the bins
        (or pixels) at which one wants to compute the model predictions. The
        X-axis is assumed to coincide with the galaxy projected major axis. The
        galaxy center is at (0,0).

        When no PSF/pixel convolution is performed (``SIGMAPSF=0`` or
        ``PIXSIZE=0``) there is a singularity at (0,0) which should be avoided
        by the input coordinates.
    YBIN:
        Vector of length ``P`` with the Y coordinates in arcseconds of the bins
        (or pixels) at which one wants to compute the model predictions. The
        Y-axis is assumed to coincide with the projected galaxy symmetry axis.

    Optional Keywords
    -----------------

    BETA:
        Vector of length ``N`` with the anisotropy
        ``beta_z = 1 - (sigma_z/sigma_R)**2`` of the individual MGE Gaussians.
    ERMS:
        Vector of length ``P`` with the 1sigma errors associated to the ``RMS``
        measurements. From the error propagation::

            ERMS = sqrt((dVel*velBin)^2 + (dSig*sigBin)^2)/RMS,

        where ``velBin`` and ``sigBin`` are the velocity and dispersion in each
        bin and ``dVel`` and ``dSig`` are the corresponding errors.
        (Default: constant errors ``ERMS=0.05*np.median(RMS)``)
    FLUX_OBS:
        Optional mean surface brightness of each bin for plotting.
    GOODBINS:
        Boolean vector of length ``P`` with values True for the bins which
        have to be included in the fit (if requested) and ``chi**2`` calculation.
        (Default: fit all bins).
    ML:
        Mass-to-light ratio to multiply the values given by SURF_POT.
        Setting this keyword is completely equivalent to multiplying the
        output ``RMSMODEL`` by ``np.sqrt(M/L)`` after the fit. This implies that
        the BH mass becomes ``MBH*(M/L)``.

        If this keyword is not set, or set to a negative number in input, the M/L
        is fitted from the data and the best-fitting M/L is returned in output.
        The BH mass of the best-fitting model is ``MBH*(M/L)``.
    NORMPSF:
        Vector of length ``Q`` with the fraction of the total PSF flux
        contained in the circular Gaussians describing the PSF of the
        observations. It has to be ``np.sum(NORMPSF) = 1``. The PSF will be used
        for seeing convolution of the model kinematics.
    NANG:
        Same as for ``NRAD``, but for the number of angular intervals
        (default: ``NANG=10``).
    NRAD:
        Number of logarithmically spaced radial positions for which the
        models is evaluated before interpolation and PSF convolution. One may
        want to increase this value if the model has to be evaluated over many
        orders of magnitutes in radius (default: ``NRAD=20``). The computation
        time scales as ``NRAD*NANG``.
    PIXANG:
        angle between the observed spaxels and the galaxy major axis X.
    PIXSIZE:
        Size in arcseconds of the (square) spatial elements at which the
        kinematics is obtained. This may correspond to the side of the spaxel
        or lenslets of an integral-field spectrograph. This size is used to
        compute the kernel for the seeing and aperture convolution.

        If this is not set, or ``PIXSIZE=0``, then convolution is not performed.
    PLOT:
        Set this keyword to produce a plot at the end of the calculation.
    QUIET:
        Set this keyword to avoid printing values on the screen.
    RBH:
        This scalar gives the sigma in arcsec of the Gaussian representing the
        central black hole of mass MBH (See Section 3.1.2 of `Cappellari 2008.
        <http://adsabs.harvard.edu/abs/2008MNRAS.390...71C>`_)
        The gravitational potential is indistinguishable from a point source
        for ``radii > 2*RBH``, so the default ``RBH=0.01`` arcsec is appropriate
        in most current situations.

        ``RBH`` should not be decreased unless actually needed!
    RMS:
        Vector of length ``P`` with the input observed velocity second moment::

            V_RMS = sqrt(velBin**2 + sigBin**2)

        at the coordinates positions given by the vectors ``XBIN`` and ``YBIN``.

        If ``RMS`` is set and ``ML`` is negative or not set, then the model is
        fitted to the data, otherwise the adopted ML is used and just the
        ``chi**2`` is returned.
    SIGMAPSF:
        Vector of length ``Q`` with the dispersion in arcseconds of the
        circular Gaussians describing the PSF of the observations.

        If this is not set, or ``SIGMAPSF=0``, then convolution is not performed.

        IMPORTANT: PSF convolution is done by creating a 2D image, with pixels
        size given by ``STEP=MAX(SIGMAPSF, PIXSIZE/2)/4``, and convolving it
        with the PSF + aperture. If the input radii RAD are very large with
        respect to STEP, the 2D image may require a too large amount of memory.
        If this is the case one may compute the model predictions at small radii
        separately from those at large radii, where PSF convolution is not
        needed.
    STEP:
        Spatial step for the model calculation and PSF convolution in arcsec.
        This value is automatically computed by default as
        ``STEP=MAX(SIGMAPSF,PIXSIZE/2)/4``. It is assumed that when ``PIXSIZE``
        or ``SIGMAPSF`` are big, high resolution calculations are not needed. In
        some cases however, e.g. to accurately estimate the central Vrms in a
        very cuspy galaxy inside a large aperture, one may want to override the
        default value to force smaller spatial pixels using this keyword.
    TENSOR:
        String specifying the component of the velocity dispersion tensor.

        ``TENSOR='xx'`` gives sigma_xx=sqrt<V_x'^2> of the component of the
        proper motion dispersion tensor in the direction parallel to the
        projected major axis.

        ``TENSOR='yy'`` gives sigma_yy=sqrt<V_y'^2> of the component of the
        proper motion dispersion tensor in the direction parallel to the
        projected symmetry axis.

        ``TENSOR='zz'`` (default) gives the usual line-of-sight V_rms=sqrt<V_z'^2>.

        ``TENSOR='xy'`` gives the mixed component <V_x'V_y'> of the proper
        motion dispersion tensor.

        ``TENSOR='xz'`` gives the mixed component <V_x'V_z'> of the proper
        motion dispersion tensor.

        ``TENSOR='yz'`` gives the mixed component <V_y'V_z'> of the proper
        motion dispersion tensor.
    VMAX:
        Maximum value of the ``Vrms`` to plot.
    VMIN:
        Minimum value of the ``Vrms`` to plot.

    Output Parameters
    -----------------

    RMSMODEL:
        Vector of length P with the model predictions for the velocity
        second moments for each bin::

            V_RMS = sqrt(vel**2 + sig**2)

        Any of the six components of the symmetric proper motion dispersion
        tensor can be provided in output using the ``TENSOR`` keyword.
    ML:
        Best fitting M/L.
    CHI2:
        Reduced ``chi**2`` describing the quality of the fit::

            chi2 = (((rms[goodBins] - rmsModel[goodBins])/erms[goodBins])**2).sum()
                 / goodBins.sum()

    FLUX:
        Vector of length ``P`` with the PSF-convolved MGE surface brightness of
        each bin in ``Lsun/pc**2``, used to plot the isophotes on the model results.

    """
    if beta is None:
        beta = np.zeros_like(surf_lum)  # Anisotropy parameter beta = 1 - (sig_z/sig_R)**2
    assert surf_lum.size == sigma_lum.size == qobs_lum.size == beta.size, "The luminous MGE components do not match"
    assert surf_pot.size == sigma_pot.size == qobs_pot.size, "The total mass MGE components do not match"
    assert xbin.size == ybin.size, "xbin and ybin do not match"
    if rms is not None:
        if erms is None:
            erms = np.full_like(rms, np.median(rms)*0.05)  # Constant ~5% errors
        if goodbins is None:
            goodbins = np.ones_like(rms, dtype=bool)
        else:
            assert goodbins.dtype == bool, "goodbins must be a boolean vector"
            assert np.any(goodbins), "goodbins must contain some True values"
        assert xbin.size == rms.size == erms.size == goodbins.size, "(rms, erms, goodbins) and (xbin, ybin) do not match"
    assert tensor in ['xx', 'yy', 'zz', 'xy', 'xz', 'yz'], \
        "`tensor` must be one of 'xx', 'yy', 'zz', 'xy', 'xz', 'yz'"

    sigmapsf = np.atleast_1d(sigmapsf)
    normpsf = np.atleast_1d(normpsf)
    assert sigmapsf.size == normpsf.size, "sigmaPSF and normPSF do not match"
    assert round(np.sum(normpsf), 2) == 1, "PSF not normalized"

    pc = distance*np.pi/0.648 # Constant factor to convert arcsec --> pc (with distance in Mpc)

    surf_lum_pc = surf_lum
    surf_pot_pc = surf_pot
    sigma_lum_pc = sigma_lum*pc         # Convert from arcsec to pc
    sigma_pot_pc = sigma_pot*pc         # Convert from arcsec to pc
    xbin_pc = xbin*pc                   # Convert all distances to pc
    ybin_pc = ybin*pc
    pixSize_pc = pixsize*pc
    sigmaPsf_pc = sigmapsf*pc
    step_pc = step*pc

    # Add a Gaussian with small sigma and the same total mass as the BH.
    # The Gaussian provides an excellent representation of the second moments
    # of a point-like mass, to 1% accuracy out to a radius 2*sigmaBH.
    # The error increses to 14% at 1*sigmaBH, independently of the BH mass.
    #
    if mbh > 0:
        sigmaBH_pc = rbh*pc # Adopt for the BH just a very small size
        surfBH_pc = mbh/(2*np.pi*sigmaBH_pc**2)
        surf_pot_pc = np.append(surfBH_pc, surf_pot_pc) # Add Gaussian to potential only!
        sigma_pot_pc = np.append(sigmaBH_pc, sigma_pot_pc)
        qobs_pot = np.append(1., qobs_pot)  # Make sure vectors do not have extra dimensions

    qobs_lum = qobs_lum.clip(0, 0.999)
    qobs_pot = qobs_pot.clip(0, 0.999)

    t = clock()
    rmsModel, psfConvolution = vrms2(
        xbin_pc, ybin_pc, inc, surf_lum_pc, sigma_lum_pc, qobs_lum,
        surf_pot_pc, sigma_pot_pc, qobs_pot, beta, tensor,
        sigmaPsf_pc, normpsf, pixSize_pc, pixang, step_pc, nrad, nang)

    if not quiet:
        print('jam_axi_rms elapsed time sec: %.2f' % (clock() - t))
        if not psfConvolution:
            txt = "No PSF convolution:"
            if np.max(sigmapsf) == 0:
                txt += " sigmapsf == 0;"
            if pixsize == 0:
                txt += " pixsize == 0;"
            print(txt)

    if tensor in ('xx', 'yy', 'zz'):
        rmsModel = np.sqrt(rmsModel.clip(0))  # Return SQRT and fix possible rounding errors

    # Analytic convolution of the MGE model with an MGE circular PSF
    # using Equations (4,5) of Cappellari (2002, MNRAS, 333, 400).
    # Broadcast triple loop over (n_MGE, n_PSF, n_bins)
    #
    sigmaX2 = sigma_lum**2 + sigmapsf[:, None]**2
    sigmaY2 = (sigma_lum*qobs_lum)**2 + sigmapsf[:, None]**2
    surf_conv = surf_lum_pc*qobs_lum*sigma_lum**2*normpsf[:, None]/np.sqrt(sigmaX2*sigmaY2)
    flux = surf_conv[..., None]*np.exp(-0.5*(xbin**2/sigmaX2[..., None] + ybin**2/sigmaY2[..., None]))
    flux = flux.sum((0, 1))  # PSF-convolved Lsun/pc**2

    if rms is None:

        chi2 = None
        if ml is None:
            ml = 1.
        else:
            rmsModel *= np.sqrt(ml)

    else:

        d, m = (rms/erms)[goodbins], (rmsModel/erms)[goodbins]
        if (ml is None) or (ml <= 0):
            ml = ((d @ m)/(m @ m))**2   # (C08 equation 51)
        rmsModel *= np.sqrt(ml)
        chi2 = ((d - m*np.sqrt(ml))**2).sum() / goodbins.sum()

        if not quiet:
            print(f'inc={inc:.1f} beta_z={beta[0]:.2f} M/L={ml:.3g} BH={mbh*ml:.2e} chi2/DOF={chi2:.3g}')
            mass = 2*np.pi*surf_pot_pc*qobs_pot*sigma_pot_pc**2
            print(f'Total mass MGE: {np.sum(mass*ml):.4g}')

        if plot:

            if flux_obs is None:
                flux_obs = flux

            rms1 = rms.copy() # Only symmetrize good bins
            rms1[goodbins] = symmetrize_velfield(xbin[goodbins], ybin[goodbins], rms[goodbins])

            if (vmin is None) or (vmax is None):
                vmin, vmax = np.percentile(rms1[goodbins], [0.5, 99.5])

            plt.clf()
            plt.subplot(121)
            plot_velfield(xbin, ybin, rms1, vmin=vmin, vmax=vmax, flux=flux_obs, **kwargs)
            plt.title(r"Input $V_{\rm rms}$")

            plt.subplot(122)
            plot_velfield(xbin, ybin, rmsModel, vmin=vmin, vmax=vmax, flux=flux, **kwargs)
            plt.plot(xbin[~goodbins], ybin[~goodbins], 'ok', mec='white')
            plt.title(r"Model $V_{\rm rms}$")
            plt.tick_params(labelleft=False)
            plt.subplots_adjust(wspace=0.03)

    return rmsModel, ml, chi2, flux

##############################################################################



"""
############################################################################

Copyright (C) 2003-2020, Michele Cappellari
E-mail: michele.cappellari_at_physics.ox.ac.uk

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your research,
I would appreciate an acknowledgement to the use of the
"mge_vcirc routine in the Python modelling package JamPy of Cappellari (2020)"

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

############################################################################

Changelog
---------

V6.0.0: MC, Oxford, 27 July 2020
++++++++++++++++++++++++++++++++

- Use new formalism and transformation of Cappellari (2020).

V5.0.0: MC, Oxford, 19 December 2018
++++++++++++++++++++++++++++++++++++

- Formatted documentation as reStructuredText.
- Removed z dependence and simplified formalism.
- Vectorized integrand: significant speedup.

V4.0.2: MC, Oxford, 17 April 2018
+++++++++++++++++++++++++++++++++

- Changed imports for JamPy as a package.
- Removed example. MC, Oxford, 17 April 2018

V4.0.1: MC, Oxford, 25 May 2014
+++++++++++++++++++++++++++++++

- Support both Python 2.7 and Python 3.

V4.0.0: MC, Oxford, 10 April 2014
+++++++++++++++++++++++++++++++++

- Translated from IDL into Python.

V3.0.2: MC, Windhoek, 1 October 2008
++++++++++++++++++++++++++++++++++++

- First released version. Included documentation. QUADVA integrator.

V3.0.0: MC, Leiden, 22 November 2005
++++++++++++++++++++++++++++++++++++

- Minor code polishing.

V3.0.0: MC, Oxford, 9 November 2006
+++++++++++++++++++++++++++++++++++

- This version retains only the few routines required for the
  computation of the circular velocity. All other unnecessary modelling
  routines have been removed.

V1.0.0: MC, Leiden, 3 February 2003
+++++++++++++++++++++++++++++++++++

- Written and tested by Michele Cappellari as part of the implementation of
  Schwarzschild's numerical orbit superposition method described in
  Cappellari et al. (2006, MNRAS).

"""

import numpy as np

from jampy.quadva import quadva

##############################################################################

def vcirc(R, dens, sigma, qintr, bhMass, soft):
    """
    Purpose
    -------

    This procedure calculates the circular velocity in the equatorial plane
    of an axisymmetric galaxy model with surface brightness described by a
    Multi-Gaussian Expansion and with a central black hole.

    It integrates equation (45) of `Cappellari (2020)
    <https://ui.adsabs.harvard.edu/abs/2020MNRAS.494.4819C>`_
    with the approach described in Section 6.2.

    Calling sequence
    ----------------

    .. code-block:: python

        from jampy.mge_vcirc import mge_vcirc

        vcirc = mge_vcirc(surf_pot, sigma_pot, qobs_pot,
                        inc_deg, mbh, distance, rad, vcirc, soft=0)

    Input parameters
    ----------------

    SURF_POT:
        vector of length M containing the peak value of the MGE Gaussians
        describing the galaxy surface density in units of Msun/pc**2 (solar
        masses per parsec**2). This is the MGE model from which the model
        potential is computed.
    SIGMA_POT:
        vector of length M containing the dispersion in arcseconds of
        the MGE Gaussians describing the galaxy surface density.
    QOBS_POT:
        vector of length M containing the observed axial ratio of the MGE
        Gaussians describing the galaxy surface density.
    INC_DEG:
        inclination in degrees (90 being edge-on).
    MBH:
        Mass of a nuclear supermassive black hole in solar masses.
    DISTANCE:
        distance of the galaxy in Mpc.
    RAD:
        Vector of length P with the radius in arcseconds, measured from the
        galaxy centre, at which one wants to compute the model predictions.

    Keywords
    --------

    SOFT:
        Softening length in arcsec for the Keplerian potential of the black
        hole. When this keyword is nonzero the black hole potential will be
        replaced by a Plummer (1911) potential with the given scale length.

    Output parameter
    ----------------

    VCIRC:
        Vector of length P with the model predictions for the circular
        velocity at the given input radii RAD.

    """

    G = 0.00430237    # (km/s)**2 pc/Msun [6.674e-11 SI units (CODATA-14)]
    scale = np.median(sigma)    # See end of Sec.6.2 of Cappellari (2020)
    R, sigma, soft = R/scale, sigma/scale, soft/scale
    q2, s2, r2, qd = qintr**2, sigma**2, R**2, qintr*dens

    integ = np.empty_like(r2)
    for j, r2j in enumerate(r2):
        integ[j] = quadva(_integrand, [-3, 3], epsabs=0, args=(r2j, qd, q2, s2))[0]

    vc = R*np.sqrt(2*np.pi*G*scale**2*integ + G*bhMass/scale*(r2 + soft**2)**(-1.5))

    return vc

##############################################################################

def mge_vcirc(surf_pc, sigma_arcsec, qobs, inc_deg, mbh, dist, rad, soft=0.):

    pc = dist*np.pi/0.648 # Constant factor to convert arcsec --> pc
    
    soft_pc = soft*pc           # Convert from arcsec to pc
    Rcirc = rad*pc              # Convert from arcsec to pc
    sigma = sigma_arcsec*pc     # Convert from arcsec to pc
    
    inc = np.radians(inc_deg)
    qintr = qobs**2 - np.cos(inc)**2
    if np.any(qintr <= 0.0):
        raise ValueError('Inclination too low for deprojection')
    qintr = np.sqrt(qintr)/np.sin(inc)                  # eq.(35) of Cappellari (2020)
    dens = surf_pc*qobs/(qintr*sigma*np.sqrt(2*np.pi))  # eq.(38) of Cappellari (2020)
    
    return vcirc(Rcirc, dens, sigma, qintr, mbh, soft_pc)

##############################################################################

def _integrand(u, r2, m, q2, s2):
    """
    Equation (45) of Cappellari (2020) with DE transformation of Section 6.2

    """
    # DE change of variables for Chandrasekhar u-integral [-3,3] -> [0,inf]
    x = np.exp(np.sinh(u)*np.pi/2)
    duds = x*np.cosh(u)*np.pi/2
    u = x[:, None]

    u1 = 1 + u
    tmp = m*np.exp(-0.5*r2/(s2*u1))/(u1**2*np.sqrt(q2 + u))

    return duds*tmp.sum(1)

##############################################################################



"""
#############################################################################

Copyright (C) 2003-2018, Michele Cappellari
E-mail: michele.cappellari_at_physics.ox.ac.uk

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your research,
I would appreciate an acknowledgement to the use of the
"JAM modelling method of Cappellari (2008)"

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

#############################################################################

MODIFICATION HISTORY:
    V1.0.0: Written and tested by Michele Cappellari, Oxford, 24 April 2008
    V1.1.0: First released version. MC, Oxford, 12 August 2008
    V1.2.0: Updated documentation. MC, Oxford, 14 August 2008
    V2.0.0: Implemented PSF convolution using interpolation on polar grid. Dramatic
        speed-up of calculation. Further documentation.
        MC, Oxford, 11 September 2008;
    V2.0.1: Bug fix: when EVEL was not given, the default was not properly set.
        Included keyword STEP. The keyword FLUX is now only used for output: the
        surface brightness for plotting is computed from the MGE model.
        MC, Windhoek, 29 September 2008
    V2.0.2: Bug fix: Velocity was not multiplied by sign(x) without convolution.
        MC, Oxford, 22 October 2008
    V2.0.3: Added keywords NRAD and NANG. Thanks to Michael Williams for reporting
        possible problems with too coarse interpolation.
        MC, Oxford, 21 November 2008
    V2.1.0: Allow for alternative definition of tangential anisotropy via the
        keyword GAMMA. MC, Oxford, 6 December 2008
    V2.1.1: Added keyword RBH. MC, Oxford 4 April 2009
    V2.1.2: Do not change KAPPA on output.
        Compute FLUX even when not plotting. MC, Oxford, 29 May 2009
    V2.1.3: Skip unnecessary interpolation when computing few points
        without PSF convolution. After feedback from Eric Emsellem.
        MC, Oxford, 6 July 2009
    V2.1.4: The routine TEST_JAM_AXISYMMETRIC_VEL with the usage example now
        adopts a more realistic input kinematics. MC, Oxford, 08 February 2010
    V2.1.5: Forces q_lum && q_pot < 1. MC, Oxford, 01 March 2010
    V2.1.6: Allow for counter-rotating components as described
        in Note 8 of the JAM paper. MC, Oxford, 10 March 2010
    V2.1.7: Use linear instead of smooth interpolation. After feedback
        from Eric Emsellem. MC, Oxford, 09 August 2010
    V2.1.8: Plot and output with FLUX keyword the PSF-convolved MGE
        surface brightness. MC, Oxford, 15 September 2010
    V2.1.9: Fix: always print output value of kappa on plots.
        MC, Oxford, 28 February 2011
    V2.2.0: Only calculates FLUX if required. MC, Oxford, 8 December 2011
    V2.2.1: Use renamed CAP_* routines to avoid potential naming conflicts.
        MC, Paranal, 8 November 2013
    V3.0.0: Translated from IDL into Python. MC, Winchester, 24 January 2014
    V3.1.0: Speed up by ~60x by broadcasting the inner integral, optimizing the
        LOS integration range and sacrificing some unnecessary accuracy.
        Included x and y proper motion components. MC, Oxford, 29 January 2014
    V3.1.1: Plot bi-symmetrized V as in IDL version. MC, Oxford, 24 February 2014
    V3.1.2: Support both Python 2.6/2.7 and Python 3.x. MC, Oxford, 25 May 2014
    V3.1.3: Use correct "MODIFICATION HISTORY". MC, Oxford, 5 August 2014
    V3.1.4: Modified final plot layout. MC, Oxford, 31 October 2014
    V3.1.5: Fixed potential program stop for models at very large radii.
        MC, Verona, 15 November 2014
    V3.1.6: Return chi2=None when vel is not given in input.
        MC, Oxford, 7 January 2015
    V3.1.7: Introduced further checks on matching input sizes.
        MC, Sydney, 5 February 2015
    V3.1.8: Improved density sampling when computing LOS integration interval.
      - Properly assign velocity symmetry before convolution as in IDL version.
      - Fixed possible program stop with negative Gaussians.
      - Changed meaning of goodbins to be a boolean vector.
        MC, Oxford, 23 May 2015
    V3.1.9: Plot bad bins on the data. MC, Oxford, 18 September 2015
    V3.1.10: Pass **kwargs for plotting. MC, Oxford, 23 February 2016
    V3.1.11: Use odd kernel size for convolution.
      - Fixed corner case with coordinates falling outside the
        interpolation region, due to finite machine precision.
        MC, Oxford, 17 February 2017
    V3.1.12: Included `flux_obs` keyword. Updated documentation.
      - Fixed DeprecationWarning in Numpy 1.12.
        MC, Oxford, 17 March 2017
    V3.1.13: Raise an error if goodbins is all False. MC, Oxford, 10 August 2017
    V3.1.14: Make default `step` depend on `sigmapsf` regardless of `pixsize`.
        MC, Oxford, 10 September 2017
    V3.1.15: Minor improvement to sampling of LOS integration range.
      - Fixed velocity symmetry with few values and no PSF convolution.
        Thanks to Martha Tabor (Nottingham) for reporting the bug.
      - Print a message when no PSF convolution was performed.
      - Broadcast kernel and MGE convolution loops.
      - Force `nsteps` to be odd. MC, Oxford, 22 January 2018
    V3.1.16: Check that PSF is normalized. MC, Oxford, 7 March 2018
    V3.1.17: Do not assume positive kappa when the input vector contains
        identical values. Thanks to Matha Tabor (Nottingham) for reporting this.
        MC, Oxford, 13 April 2018
    V3.1.18: Fixed MatplotlibDeprecationWarning in Matplotlib 2.2.
      - Changed imports for jam as a package.
      - Removed example. MC, Oxford, 17 April 2018
    V3.1.19: Dropped Python 2.7 support. MC, Oxford, 12 May 2018
    V3.1.20: Fixed clock DeprecationWarning in Python 3.7.
        MC, Oxford, 27 September 2018
    V3.2.0: Replaced quadva with TAHN LOS quadrature (~4x speedup).
        MC, Oxford, 30 May 2019
    V3.2.1: Use analytic mge_surf in convolution. MC, Oxford, 31 October 2019
    V3.2.2: Invert sign of V_x for consistency with Cappellari (2020) and
        with the new jam_axi_proj. MC, Oxford, 22 July 2020

"""

import numpy as np
import matplotlib.pyplot as plt
from scipy import special, signal, ndimage, integrate
from time import perf_counter as clock

from plotbin.plot_velfield import plot_velfield
from plotbin.symmetrize_velfield import symmetrize_velfield

##############################################################################

def mge_surf(x, y, surf, sigma, qobs):
    """ MGE surface brightness for a set of coordinates (x, y) """

    mge = surf*np.exp(-0.5/sigma**2*(x[..., None]**2 + (y[..., None]/qobs)**2))

    return mge.sum(-1)

##############################################################################

def bilinear_interpolate(xv, yv, im, xout, yout):
    """
    The input array has size im[ny,nx] as in the output
    of im = f(meshgrid(xv, yv))
    xv and yv are vectors of size nx and ny respectively.

    """
    ny, nx = im.shape
    assert (nx, ny) == (xv.size, yv.size), "Input arrays dimensions do not match"

    xi = (nx-1.)/(xv[-1] - xv[0]) * (xout - xv[0])
    yi = (ny-1.)/(yv[-1] - yv[0]) * (yout - yv[0])

    return ndimage.map_coordinates(im.T, [xi, yi], order=1, mode='nearest')

##############################################################################

def rotate_points(x, y, ang):
    """
    Rotates points conter-clockwise by an angle ANG in degrees.
    Michele cappellari, Paranal, 10 November 2013
    """
    theta = np.radians(ang)
    xNew = x*np.cos(theta) - y*np.sin(theta)
    yNew = x*np.sin(theta) + y*np.cos(theta)

    return xNew, yNew

##############################################################################

def integrand(u, s2_lum, q2_lum, dens_pot, s2_pot, q2_pot,
               R2, z2, bani, kappa, gamma, nu):

    """
    This routine computes the inner integrand of Eq.(38) of Cappellari (2008)
    for a model with constant anisotropy
    sigma_R**2 = b*sigma_z**2 and <V_R*V_z> = 0.

    """
    u2 = u**2
    s2q2_lum = s2_lum*q2_lum
    e2_pot = 1. - q2_pot

    c = e2_pot - s2q2_lum/s2_pot                            # equation (22)
    d = 1 - bani*q2_lum - ((1 - bani)*c + e2_pot*bani)*u2   # equation (23)

    # The next line gives an extra term for equation (38) using the alternative
    # definition of tangential anisotropy of equation (33) instead of the
    # default form of equation (35). This term is zero by default (gamma=0)
    # (also see note 10 here http://adsabs.harvard.edu/abs/2015ApJ...804L..21C)
    if np.any(gamma):
        d = d + gamma*bani*s2q2_lum/R2

    # Double summation over (j, k) of eq.(38) for all values of integration
    # variable u and z' LOS positions. The quadruple loop in (j, k, u, z') is
    # replaced by broadcast Numpy array operations.
    p2 = 1 - e2_pot*u2
    hj = np.exp(-0.5/s2_pot*u2*(R2+z2/p2))/np.sqrt(p2)      # equation (17)
    e = np.sqrt(q2_pot)*dens_pot*u2*hj
    arr = np.sign(kappa)*kappa**2*nu*(d/(1 - c*u2))*e       # equation (38)

    # arr has the dimensions (q_lum.size, q_pot.size, u.size, z.size)

    return arr.sum((0, 1)) # sum over light and mass Gaussians

##############################################################################

def los_integrand(z1, dens_lum, sigma_lum, qintr_lum,
                   dens_pot, sigma_pot, qintr_pot,
                   x1, y1, inc, beta, kappa, gamma, component, nsteps):
    """
    This function returns the outer integrand in equation (38)
    of Cappellari (2008, hereafter C08). Also included is the optional
    extension to proper motions as in D'Souza & Rix (2013, hereafter DR13).

    """
    s2_lum = sigma_lum**2
    q2_lum = qintr_lum**2
    s2_pot = sigma_pot**2
    q2_pot = qintr_pot**2
    bani = 1./(1. - beta) # Anisotropy ratio b = (sig_R/sig_z)**2

    # Tracer distribution parameters
    dens_lum = dens_lum[:, None, None, None]
    s2_lum = s2_lum[:, None, None, None]
    q2_lum = q2_lum[:, None, None, None]
    bani = bani[:, None, None, None]
    kappa = kappa[:, None, None, None]
    gamma = gamma[:, None, None, None]

    # Mass distribution parameters
    dens_pot = dens_pot[None, :, None, None]
    s2_pot = s2_pot[None, :, None, None]
    q2_pot = q2_pot[None, :, None, None]

    uu = np.linspace(0., 1., nsteps)  # integration variable from 0-1
    u = uu[None, None, :, None]

    z1 = z1[None, None, None, :]  # LOS integration variable
    R2 = (z1*np.sin(inc) - y1*np.cos(inc))**2 + x1**2  # Equation (25) of C08
    z2 = (z1*np.cos(inc) + y1*np.sin(inc))**2

    nu = dens_lum*np.exp(-0.5/s2_lum*(R2 + z2/q2_lum))
    tmp = integrand(u, s2_lum, q2_lum, dens_pot, s2_pot, q2_pot,
                     R2, z2, bani, kappa, gamma, nu)
    tmp = integrate.simps(tmp, dx=uu[1], axis=0)

    nu = nu.sum((0, 1, 2))  # Sum over luminous Gaussians

    if np.ptp(kappa):
        integ = np.sign(tmp)*np.sqrt(abs(nu*tmp))  # Note 8 pg. 77 of C08
    else:
        integ = np.sqrt(nu*tmp.clip(0)) if kappa[0] > 0 else -np.sqrt((-nu*tmp).clip(0))

    if component == 'x':
        f = z1*np.sin(inc) - y1*np.cos(inc)     # See Eq.(B9) of DR13
    elif component == 'y':
        f = x1*np.cos(inc)                      # See Eq.(B8) of DR13
    elif component == 'z':
        f = x1*np.sin(inc)                      # z' LOS equation (38) of C08

    # This routine returns a vector of values computed at different values of z1
    #
    G = 0.004301    # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA-14)]

    return 2.*np.sqrt(np.pi*G)*f*integ

##############################################################################

def surf_vlos(x1, y1, inc,
               dens_lum, sigma_lum, qintr_lum,
               dens_pot, sigma_pot, qintr_pot,
               beta, kappa, gamma, component, nsteps):
    """
    This routine gives the projected weighted first moment Sigma*<V_los>

    """
    # TANH quadrature rule
    rmax = 3*np.max(sigma_lum)
    tmax = 8    # break is rmax/tmax
    t, dt = np.linspace(-tmax, tmax, nsteps, retstep=True)
    scale = rmax/np.sinh(tmax)
    z1 = scale*np.sinh(t)
    dxdt = dt*scale*np.cosh(t)

    sb_mu1 = np.empty_like(x1)
    for j, (xj, yj) in enumerate(zip(x1, y1)):
        integ = los_integrand(z1, dens_lum, sigma_lum, qintr_lum,
                               dens_pot, sigma_pot, qintr_pot,
                               xj, yj, inc, beta, kappa, gamma, component, nsteps)
        sb_mu1[j] = integ @ dxdt     # DE quadrature

    return sb_mu1

##############################################################################

def vlos(x, y, inc_deg,
         surf_lum, sigma_lum, qobs_lum,
         surf_pot, sigma_pot, qobs_pot,
         beta, kappa, gamma, sigmaPsf, normPsf,
         pixSize, pixAng, step, nrad, nang, component, nsteps):
    """
    This routine gives the second V moment after convolution with a PSF.
    The convolution is done unp.sing interpolation of the model on a
    polar grid, as described in Appendix A of Cappellari (2008).

    """
    # Axisymmetric deprojection of both luminous and total mass.
    # See equation (12)-(14) of Cappellari (2008)
    #
    inc = np.radians(inc_deg)

    qintr_lum = qobs_lum**2 - np.cos(inc)**2
    if np.any(qintr_lum <= 0):
        raise RuntimeError('Inclination too low q_lum < 0')
    qintr_lum = np.sqrt(qintr_lum)/np.sin(inc)
    if np.any(qintr_lum < 0.05):
        raise RuntimeError('q_lum < 0.05 components')
    dens_lum = surf_lum*qobs_lum / (sigma_lum*qintr_lum*np.sqrt(2*np.pi))

    qintr_pot = qobs_pot**2 - np.cos(inc)**2
    if np.any(qintr_pot <= 0):
        raise RuntimeError('Inclination too low q_pot < 0')
    qintr_pot = np.sqrt(qintr_pot)/np.sin(inc)
    if np.any(qintr_pot < 0.05):
        raise RuntimeError('q_pot < 0.05 components')
    dens_pot = surf_pot*qobs_pot / (sigma_pot*qintr_pot*np.sqrt(2*np.pi))

    # Define parameters of polar grid for interpolation
    #
    w = sigma_lum < np.max(np.abs(x)) # Characteristic MGE axial ratio in observed range

    if w.sum() < 3:
        qmed = np.median(qobs_lum)
    else:
        qmed = np.median(qobs_lum[w])

    rell = np.sqrt(x**2 + (y/qmed)**2) # Elliptical radius of input (x,y)

    psfConvolution = (np.max(sigmaPsf) > 0) and (pixSize > 0)

    # Kernel step is 1/4 of largest value between sigma(min) and 1/2 pixel side.
    # Kernel half size is the sum of 3*sigma(max) and 1/2 pixel diagonal.
    if (nrad*nang > x.size) and (not psfConvolution): # Just calculate values

        xPol = x
        yPol = y

    else:  # Interpolate values on polar grid

        if psfConvolution:   # PSF convolution
            if step == 0:
                step = np.min(sigmaPsf)/4.
            mx = 3*np.max(sigmaPsf) + pixSize/np.sqrt(2)
        else:                                   # No convolution
            step = max(np.sqrt(2), np.min(rell)) # Minimum radius of 1pc
            mx = 0

        # Make linear grid in log of elliptical radius RAD and eccentric anomaly ANG
        # See Appendix A
        rmax = np.max(rell) + mx # Major axis of ellipse containing all data + convolution
        logRad = np.linspace(np.log(step/np.sqrt(2)), np.log(rmax), nrad) # Linear grid in np.log(rell)
        ang = np.linspace(0, np.pi/2, nang) # Linear grid in eccentric anomaly
        radGrid, angGrid = map(np.ravel, np.meshgrid(np.exp(logRad), ang))
        xPol = radGrid*np.cos(angGrid)
        yPol = radGrid*np.sin(angGrid) * qmed

    # The model Vel computation is only performed on the polar grid
    # which is then used to interpolate the values at any other location
    wmPol = surf_vlos(xPol, yPol, inc,
                      dens_lum, sigma_lum, qintr_lum,
                      dens_pot, sigma_pot, qintr_pot,
                      beta, kappa, gamma, component, nsteps)

    if psfConvolution: # PSF convolution

        nx = int(np.ceil(rmax/step))
        ny = int(np.ceil(rmax*qmed/step))
        x1 = np.linspace(0.5 - nx, nx - 0.5, 2*nx)*step
        y1 = np.linspace(0.5 - ny, ny - 0.5, 2*ny)*step
        xCar, yCar = np.meshgrid(x1, y1)  # Cartesian grid for convolution

        # Interpolate MGE model and Vel over cartesian grid
        r1 = 0.5*np.log(xCar**2 + (yCar/qmed)**2)  # Log elliptical radius of cartesian grid
        e1 = np.arctan2(np.abs(yCar/qmed), np.abs(xCar)) # Eccentric anomaly of cartesian grid

        # Division by mgePol before interpolation reduces interpolation error
        mgeCar = mge_surf(xCar, yCar, surf_lum, sigma_lum, qobs_lum)
        mgePol = mge_surf(xPol, yPol, surf_lum, sigma_lum, qobs_lum)
        wmCar = mgeCar*bilinear_interpolate(logRad, ang, (wmPol/mgePol).reshape(nang, nrad), r1, e1)

        if component in ('y', 'z'):
            wmCar *= np.sign(xCar) # Calculation was done in positive quadrant
        else:
            wmCar *= np.sign(yCar)  # Calculation was done in positive quadrant

        nk = int(np.ceil(mx/step))
        kgrid = np.linspace(-nk, nk, 2*nk + 1)*step
        xgrid, ygrid = np.meshgrid(kgrid, kgrid) # Kernel is square
        if pixAng != 0:
            xgrid, ygrid = rotate_points(xgrid, ygrid, pixAng)

        # Compute kernel with equation (A6) of Cappellari (2008).
        # Normalization is irrelevant here as it cancels out.
        #
        dx = pixSize/2
        sp = np.sqrt(2)*sigmaPsf
        xg, yg = xgrid[..., None], ygrid[..., None]
        kernel = normPsf*(special.erf((dx - xg)/sp) + special.erf((dx + xg)/sp)) \
                        *(special.erf((dx - yg)/sp) + special.erf((dx + yg)/sp))
        kernel = kernel.sum(2)   # Sum over PSF components

        # Seeing and aperture convolution with equation (A3)
        #
        muCar = signal.fftconvolve(wmCar, kernel, mode='same') \
              / signal.fftconvolve(mgeCar, kernel, mode='same')

        # Interpolate convolved image at observed apertures.
        # Aperture integration was already included in the kernel.
        #
        mu = bilinear_interpolate(x1, y1, muCar, x, y)

    else: # No PSF convolution

        muPol = wmPol/mge_surf(xPol, yPol, surf_lum, sigma_lum, qobs_lum)

        if nrad*nang > x.size:      # Just returns values
            mu = muPol
        else:                      # Interpolate values
            r1 = 0.5*np.log(x**2 + (y/qmed)**2) # Log elliptical radius of input (x,y)
            e1 = np.arctan2(np.abs(y/qmed), np.abs(x))    # Eccentric anomaly of input (x,y)
            mu = bilinear_interpolate(logRad, ang, muPol.reshape(nang, nrad), r1, e1)

            if component in ('y', 'z'):
                mu *= np.sign(x)  # Calculation was done in positive quadrant
            else:
                mu *= np.sign(y)  # Calculation was done in positive quadrant

    return mu, psfConvolution

##############################################################################

def jam_axi_vel(surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
                inc, mbh, distance, xbin, ybin, beta=None, component='z',
                evel=None, flux_obs=None, gamma=None, goodbins=None,
                kappa=None, nang=10, normpsf=1., nrad=20, nsteps=65, pixang=0.,
                pixsize=0., plot=True, quiet=False, rbh=0.01, sigmapsf=0.,
                step=0., vel=None, vmax=None, **kwargs):

    """

    Purpose
    -------

    This procedure calculates a prediction for the mean line-of-sight velocity
    V_z and the two proper motion components (V_x, V_y) for an anisotropic
    axisymmetric galaxy model. It implements the solution of the anisotropic
    Jeans equations presented in `Cappellari (2008, MNRAS, 390, 71)
    <https://ui.adsabs.harvard.edu/abs/2008MNRAS.390...71C>`_
    PSF convolution is done as described in the Appendix of that paper.

    Calling sequence
    ----------------

    .. code-block:: python

        from jampy.jam_axi_vel import jam_axi_vel

        velModel, kappa, chi2, flux = jam_axi_vel(
            surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
            inc, mbh, distance, xbin, ybin, beta=None, component='z',
            evel=None, gamma=None, goodbins=None, kappa=None, nang=10,
            normpsf=1., nrad=20, nsteps=50, pixang=0., pixsize=0.,
            plot=True, quiet=False, rbh=0.01, sigmapsf=0., step=0.,
            vel=None, vmax=None)

    Input parameters
    ----------------

    SURF_LUM:
        vector of length N containing the peak surface brightness of the
        MGE Gaussians describing the galaxy surface brightness in units of
        Lsun/pc^2 (solar luminosities per parsec^2).
    SIGMA_LUM:
        vector of length N containing the dispersion in arcseconds of
        the MGE Gaussians describing the galaxy surface brightness.
    QOBS_LUM:
        vector of length N containing the observed axial ratio of the MGE
        Gaussians describing the galaxy surface brightness.
    SURF_POT:
        vector of length M containing the peak value of the MGE Gaussians
        describing the galaxy surface density in units of Msun/pc^2 (solar
        masses per parsec^2). This is the MGE model from which the model
        potential is computed.

        In a common usage scenario, with a self-consistent model, one will use
        the same Gaussians for both the surface brightness and the potential.
        One will use JAM_AXISYMMETRIC_RMS to fit the global M/L and then call
        JAM_AXISYMMETRIC_VEL with SURF_POT = SURF_LUM*(M/L),
        SIGMA_POT = SIGMA_LUM and QOBS_POT = QOBS_LUM.
    SIGMA_POT:
        vector of length M containing the dispersion in arcseconds of
        the MGE Gaussians describing the galaxy surface density.
    QOBS_POT:
        vector of length M containing the observed axial ratio of the MGE
        Gaussians describing the galaxy surface density.
    INC:
        inclination in degrees (90 being edge-on).
    MBH:
        Mass of a nuclear supermassive black hole in solar masses.
    DISTANCE:
        distance in Mpc.
    XBIN:
        Vector of length P with the X coordinates in arcseconds of the bins
        (or pixels) at which one wants to compute the model predictions. The
        X-axis is assumed to coincide with the galaxy projected major axis. The
        galaxy center is at (0,0).

        When no PSF/pixel convolution is performed (SIGMAPSF=0 or PIXSIZE=0)
        there is a singularity at (0,0) which should be avoided by the input
        coordinates.
    YBIN:
        Vector of length P with the Y coordinates in arcseconds of the bins
        (or pixels) at which one wants to compute the model predictions. The
        Y-axis is assumed to concide with the projected galaxy symmetry axis.

    Optional keywords
    -----------------

    BETA:
        Vector of length N with the vertical anisotropy
        beta_z = 1 - (sigma_z/sigma_R)^2 of the individual MGE Gaussians.
        A scalar can be used if the model has constant anisotropy.
        (Default: BETA=0)
    COMPONENT:
        Set this keyword to either 'x' (=major axis), 'y' (=minor axis)
        or 'z' (=line-of-sight: default) to calculate the corresponding component
        of the velocity. The extension for x and y components is done as in
        D'Souza & Rix (2013, MNRAS, 429, 1887).
    EVEL:
        Vector of length P with the 1sigma error associated to the VEL
        measurements in each bin. (Default EVEL=1)
    FLUX_OBS:
        Optional mean surface brigfhtness of each bin for plotting.
    GAMMA:
        Vector of length N with the tangential anisotropy
        gamma = 1 - (sigma_phi/sigma_R)^2 of the individual MGE Gaussians.
        A scalar can be used if the model has constant anisotropy.
        (Default: GAMMA=0)

        IMPORTANT: In a common usage this keyword is *not* needed and the
        tangential anisotropy is only parameterized with the much more
        efficient KAPPA parameter below. However this form, which uses the
        definition of tangential anisotropy of equation (33) of Cappellari
        (2008), can be sometimes useful as an alternative (Sec.3.1.5 of
        Cappellari 2008 for a discussion).

        When this keyword is set and nonzero, one should generally not set the
        input keyword VEL with the vector of observed velocities. In this way
        KAPPA=1 is not fitted and the output VELMODEL has a well defined shape
        of the velocity ellipsoid. However VEL can still be set and KAPPA can
        be fitted to allow for further generality.
    GOODBINS:
        Boolean vector of length P with values True for the bins which
        have to be included in the fit (if requested) and chi^2 calculation.
        (Default: fit all bins).
    KAPPA:
        Rotation parameter (Default: KAPPA=1). Vector of length N defining
        by how much the model rotation of each individual MGE Gaussian is
        scaled, with respect to the velocity of a model with an oblate (GAMMA=0)
        velocity ellipsoid (Sec.3.1.5 of Cappellari 2008 for details).

        If the observed velocities are passed via the keyword VEL then KAPPA
        is the one required to match the observed projected angular momentum
        via equation (52) of Cappellari (2008).

        If the input KAPPA is a vector of length N, then the full vector is
        scaled to match the observed projected angular momentum.

        KAPPA can have both positive and negative elements to model
        counter-rotating stellar components (e.g. the S0 galaxy NGC4550).

        If GAMMA is set and nonzero then KAPPA defines by how much the model
        velocity field is scaled with respect to a model with tangential
        anisotropy GAMMA.
    NANG:
        Same as for NRAD, but for the number of angular intervals
        (default: NANG=10).
    NORMPSF:
        Vector of length Q with the fraction of the total PSF flux
        contained in the circular Gaussians describing the PSF of the
        observations. It has to be total(NORMPSF) = 1. The PSF will be used for
        seeing convolution of the model kinematics.
    NRAD:
        Number of logarithmically spaced radial positions for which the
        models is evaluated before interpolation and PSF convolution. One may
        want to increase this value if the model has to be evaluated over many
        orders of magnitudes in radius (default: NRAD=20). The computation time
        scales as NRAD*NANG.
    NSTEPS:
        Steps for the integration in the interval [0, 1] (default 50).
    PIXANG:
        angle between the observed spaxels and the galaxy major axis X.
    PIXSIZE:
        Size in arcseconds of the (square) spatial elements at which the
        kinematics is obtained. This may correspond to the side of the spaxel
        or lenslets of an integral-field spectrograph. This size is used to
        compute the kernel for the seeing and aperture convolution.

        If this is not set, or PIXSIZE=0, then convolution is not performed.
    PLOT:
        Set this keyword to produce a plot at the end of the calculation.
    QUIET:
        Set this keyword not to print values on the screen.
    RBH:
        This scalar gives the sigma in arcsec of the Gaussian representing the
        central black hole of mass MBH (See Section 3.1.2 of Cappellari 2008).
        The gravitational potential is indistinguishable from a point source
        for radii > 2*RBH, so the default RBH=0.01 arcsec is appropriate in
        most current situations.

        RBH should not be decreased unless actually needed!
    SIGMAPSF:
        Vector of length Q with the dispersion in arcseconds of the
        circular Gaussians describing the PSF of the observations.

        If this is not set, or SIGMAPSF=0, then convolution is not performed.

        IMPORTANT: PSF convolution is done by creating a 2D image, with pixels
        size given by STEP=MAX(SIGMAPSF,PIXSIZE/2)/4, and convolving it with
        the PSF + aperture. If the input radii RAD are very large with respect
        to STEP, the 2D image may require a too large amount of memory. If this
        is the case one may compute the model predictions at small radii
        separately from those at large radii, where PSF convolution is not
        needed.
    STEP:
        Spatial step for the model calculation and PSF convolution in arcsec.
        This value is automatically computed by default as
        STEP=MAX(SIGMAPSF,PIXSIZE/2)/4. It is assumed that when PIXSIZE or
        SIGMAPSF are big, high resolution calculations are not needed. In some
        cases however, e.g. to accurately estimate the central Vrms in a very
        cuspy galaxy inside a large aperture, one may want to override the
        default value to force smaller spatial pixels using this keyword.
    VEL:
        Vector of length P with the input observed mean stellar velocity V at
        the coordinates positions given by the vectors XBIN and YBIN.
    VMAX:
        Maximum absolute value of the velocity to plot.

    Output parameters
    -----------------

    VELMODEL:
        Vector of length P with the model predictions for the mean
        velocity of each bin.
    KAPPA:
        Best fitting kappa parameter. If kappa is given in input, this value
        represents by how much the input need to be scaled to reproduce the
        projected angular momentum of the data.
    CHI2:
        Reduced chi^2 describing the quality of the fit::

            chi^2 = total(((vel[goodBins]-velModel[goodBins])/evel[goodBins])^2)
                  / goodBins.sum()

        IMPORTANT: by default the rotation parameter KAPPA is determined by
        requiring the projected angular momentum to be the same in the data and
        in the model. The chi^2 is not necessarily minimized (Sec.4.2 of
        Cappellari 2008 for details).
    FLUX:
        Vector of length P with the PSF-convolved MGE surface brightness of
        each bin in Lsun/pc^2, used to plot the isophotes on the model results.

    """
    assert nsteps % 2 == 1, "`nsteps` must be odd"  # for optimal Simpson
    if beta is None:
        beta = np.zeros_like(surf_lum)  # Anisotropy parameter beta = 1 - (sig_z/sig_R)**2  (beta=0-->circle)
    if kappa is None:
        kappa = np.ones_like(surf_lum)  # Anisotropy parameter: V_obs = kappa * V_iso (kappa=1-->circle)
    if gamma is None:
        gamma = np.zeros_like(surf_lum)  # Anisotropy parameter gamma = 1 - (sig_phi/sig_R)^2 (gamma=0-->circle)
    assert (surf_lum.size == sigma_lum.size == qobs_lum.size
            == beta.size == gamma.size == kappa.size), "The luminous MGE components and anisotropies do not match"
    assert surf_pot.size == sigma_pot.size == qobs_pot.size, "The total mass MGE components do not match"
    assert xbin.size == ybin.size, "xbin and ybin do not match"
    if vel is not None:
        if evel is None:
            evel = np.full_like(vel, 5.)  # Constant 5 km/s errors
        if goodbins is None:
            goodbins = np.ones_like(vel, dtype=bool)
        else:
            assert goodbins.dtype == bool, "goodbins must be a boolean vector"
            assert np.any(goodbins), "goodbins must contain some True values"
        assert xbin.size == vel.size == evel.size == goodbins.size, "(vel, evel, goodbins) and (xbin, ybin) do not match"
    assert component in ['x', 'y', 'z'], "component must be one of 'x', 'y', 'z'"

    sigmapsf = np.atleast_1d(sigmapsf)
    normpsf = np.atleast_1d(normpsf)
    assert sigmapsf.size == normpsf.size, "sigmaPSF and normPSF do not match"
    assert round(np.sum(normpsf), 2) == 1, "PSF not normalized"

    pc = distance*np.pi/0.648 # Constant factor to convert arcsec --> pc

    surf_lum_pc = surf_lum
    surf_pot_pc = surf_pot
    sigma_lum_pc = sigma_lum*pc         # Convert from arcsec to pc
    sigma_pot_pc = sigma_pot*pc         # Convert from arcsec to pc
    xbin_pc = xbin*pc                   # Convert all distances to pc
    ybin_pc = ybin*pc
    pixSize_pc = pixsize*pc
    sigmaPsf_pc = sigmapsf*pc
    step_pc = step*pc

    # Add a Gaussian with small sigma and the same total mass as the BH.
    # The Gaussian provides an excellent representation of the second moments
    # of a point-like mass, to 1% accuracy out to a radius 2*sigmaBH.
    # The error increses to 14% at 1*sigmaBH, independently of the BH mass.
    #
    if mbh > 0:
        sigmaBH_pc = rbh*pc # Adopt for the BH just a very small size
        surfBH_pc = mbh/(2*np.pi*sigmaBH_pc**2)
        surf_pot_pc = np.append(surfBH_pc, surf_pot_pc) # Add Gaussian to potential only!
        sigma_pot_pc = np.append(sigmaBH_pc, sigma_pot_pc)
        qobs_pot = np.append(1., qobs_pot)  # Make sure vectors do not have extra dimensions

    qobs_lum = qobs_lum.clip(0, 0.999)
    qobs_pot = qobs_pot.clip(0, 0.999)

    t = clock()
    velModel, psfConvolution = vlos(
        xbin_pc, ybin_pc, inc, surf_lum_pc, sigma_lum_pc, qobs_lum,
        surf_pot_pc, sigma_pot_pc, qobs_pot, beta, kappa, gamma,
        sigmaPsf_pc, normpsf, pixSize_pc, pixang, step_pc,
        nrad, nang, component, nsteps)

    if not quiet:
        print('jam_axis_vel elapsed time sec: %.2f' % (clock() - t))
        if not psfConvolution:
            txt = "No PSF convolution:"
            if np.max(sigmapsf) == 0:
                txt += " sigmapsf == 0;"
            if pixsize == 0:
                txt += " pixsize == 0;"
            print(txt)

    # Analytic convolution of the MGE model with an MGE circular PSF
    # using Equations (4,5) of Cappellari (2002, MNRAS, 333, 400).
    # Broadcast triple loop over (n_MGE, n_PSF, n_bins)
    #
    sigmaX2 = sigma_lum**2 + sigmapsf[:, None]**2
    sigmaY2 = (sigma_lum*qobs_lum)**2 + sigmapsf[:, None]**2
    surf_conv = surf_lum_pc*qobs_lum*sigma_lum**2*normpsf[:, None]/np.sqrt(sigmaX2*sigmaY2)
    flux = surf_conv[..., None]*np.exp(-0.5*(xbin**2/sigmaX2[..., None] + ybin**2/sigmaY2[..., None]))
    flux = flux.sum((0, 1))  # PSF-convolved Lsun/pc**2

    ####### Output and optional M/L fit
    # If RMS keyword is not given all this section is skipped

    if vel is None:
        
        chi2 = None
        
    else:

        # Only consider the good bins for the chi**2 estimation
        #
        # Scale by having the same angular momentum
        # in the model and in the galaxy (equation 52)
        #
        kappa2 = np.sum(np.abs(vel[goodbins]*xbin[goodbins])) \
               / np.sum(np.abs(velModel[goodbins]*xbin[goodbins]))
        kappa = kappa*kappa2  # Rescale input kappa by the best fit

        # Measure the scaling one would have from a standard chi^2 fit of the V field.
        # This value is only used to get proper sense of rotation for the model.
        # y1 = rms; dy1 = erms (y1 are the data, y2 the model)
        # scale = total(y1*y2/dy1^2)/total(y2^2/dy1^2)  (equation 51)
        #
        kappa3 = np.sum(vel[goodbins]*velModel[goodbins]/evel[goodbins]**2) \
               / np.sum((velModel[goodbins]/evel[goodbins])**2)

        velModel *= kappa2*np.sign(kappa3)

        chi2 = np.sum(((vel[goodbins]-velModel[goodbins])/evel[goodbins])**2) / goodbins.sum()

        if not quiet:
            print('inc=%.1f kappa=%.2f beta_z=%.2f BH=%.2e chi2/DOF=%.3g' % (inc, kappa[0], beta[0], mbh, chi2))
            mass = 2*np.pi*surf_pot_pc*qobs_pot*sigma_pot_pc**2
            print('Total mass MGE %.4g:' % np.sum(mass))

        if plot:

            if flux_obs is None:
                flux_obs = flux

            vel1 = vel.copy() # Only symmetrize good bins
            vel1[goodbins] = symmetrize_velfield(xbin[goodbins], ybin[goodbins], vel[goodbins], sym=1)

            if vmax is None:
                vmax = np.percentile(np.abs(vel1[goodbins]), 99)

            plt.clf()
            plt.subplot(121)
            plot_velfield(xbin, ybin, vel1, vmin=-vmax, vmax=vmax, flux=flux_obs, **kwargs)
            plt.title(r"Input $V_{\rm los}$")

            plt.subplot(122)
            plot_velfield(xbin, ybin, velModel, vmin=-vmax, vmax=vmax, flux=flux, **kwargs)
            plt.plot(xbin[~goodbins], ybin[~goodbins], 'ok', mec='white')
            plt.title(r"Model $V_{\rm los}$")
            plt.tick_params(labelleft=False)
            plt.subplots_adjust(wspace=0.03)

    return velModel, kappa, chi2, flux

##############################################################################



#!/usr/bin/env python
"""
    Copyright (C) 2019-2020, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from my web page
    http://purl.org/cappellari/software

CHANGELOG
---------

V1.0.1: MC, Oxford, 21 April 2020
    - Made a separate file

V1.0.0: Michele Cappellari, Oxford, 08 November 2019
    - Written and tested

"""
from os import path
import numpy as np
import matplotlib.pyplot as plt

import jampy as jampy_package
from jampy.jam_axi_intr import jam_axi_intr

##############################################################################

def satoh_solution(R, z, beta):
    """
    Implements the analytic Jeans solution of eq.(68)-(69) from
    `Cappellari (2020) <https://ui.adsabs.harvard.edu/abs/2020MNRAS.494.4819C>`_

    """
    G = 0.004301  # (km/s)^2 pc/Msun [6.674e-11 SI units (CODATA-14)]
    a = b = M = 1

    Q = np.sqrt(b**2 + z**2)
    S = np.sqrt(a**2 + 2*a*Q + R**2 + z**2)
    sig2z = G*M*Q*(a + 2*Q)/(3*Q*(a + 2*Q) + S**2)/(2*S)        # C20 eq.(68)
    v2phi = sig2z*(1 - 6*(R/S)**2)/(1 - beta) + G*M*R**2/S**3   # C20 eq.(69)

    return sig2z, v2phi

##############################################################################

def jam_axi_intr_example():
    """
    Usage example for jam_axi_intr().
    It takes about 1s on a 3GHz CPU.
    I use as "data" the anisotropic cylindrical solution
    from Cappellari (2020, MNRAS) for the density by Satoh (1980).
    The MGE fit is from Table 1 of the above paper
    https://ui.adsabs.harvard.edu/abs/2020MNRAS.494.4819C

    """
    rr = np.linspace(0, 16, 32)
    zz = np.linspace(0, 6, 12)
    Rbin, zbin = map(np.ravel, np.meshgrid(rr, zz))
    w = (Rbin == 0) & (zbin == 0)
    Rbin, zbin = Rbin[~w], zbin[~w]  # Remove central singularity
    jampy_dir = path.dirname(path.realpath(jampy_package.__file__))
    log_dens_lum, log_sigma_lum, qintr_lum = np.loadtxt(jampy_dir + '/examples/cappellari2020_table1.txt', unpack=True)
    dens_lum, sigma_lum = 10**log_dens_lum, 10**log_sigma_lum

    beta0 = -0.78
    sig2z, v2phi = satoh_solution(Rbin, zbin, beta0)

    gamma0 = 0.78
    beta = np.full_like(dens_lum, beta0)
    gamma = np.full_like(dens_lum, gamma0)

    dens_pot = dens_lum
    sigma_pot = sigma_lum
    qintr_pot = qintr_lum

    mbh = 0
    sig2R = sig2z/(1 - beta0)
    sig2phi = sig2R*(1- gamma0)
    data = np.sqrt([sig2R, sig2z, sig2phi, v2phi])
    errors = np.full_like(data, 5.28e-5)  # make chi2 = 1

    jam = jam_axi_intr(dens_lum, sigma_lum, qintr_lum, dens_pot, sigma_pot, qintr_pot,
                        mbh, Rbin, zbin, beta=beta, gamma=gamma, align='cyl',
                        plot=1, data=data, errors=errors, ml=None, nodots=1)

    print(f"M/L = {jam.ml:.3f}; chi2 = {jam.chi2:.3f}")

    plt.pause(1)

##############################################################################

if __name__ == '__main__':

    jam_axi_intr_example()



#!/usr/bin/env python

"""
V1.0.0: Michele Cappellari, Oxford, 17 April 2018

"""

import numpy as np
import matplotlib.pyplot as plt

from jampy.legacy.jam_sph_rms import jam_sph_rms

def jam_sph_rms_example():
    """
    Usage example for jam_sph_rms().
    It takes about 1s on a 2.5 GHz computer

    """

    # Realistic MGE galaxy surface brightness.
    # The surface brightness is in L_sun/pc^2 and the sigma in arcsec
    #
    surf_pc = np.array([6229., 3089., 5406., 8443., 4283., 1927., 708.8, 268.1, 96.83])
    sigma_arcsec = np.array([0.0374, 0.286, 0.969, 2.30, 4.95, 8.96, 17.3, 36.9, 128.])

    # Realistic observed stellar kinematics. It comes from AO observations
    # at R<2" and seeing-limited long slit observations at larger radii.
    # The galaxy has negligible rotation and we can use sigma as V_RMS
    #
    sig = np.array([395., 390., 387., 385., 380., 365., 350., 315., 310., 290., 260.])  # km/s
    erms = sig*0.02  # assume 2% errors in sigma
    rad = np.array([0.15, 0.2, 0.3, 0.4, 0.5, 1, 1.5, 3, 5, 9, 15])  # arcsec

    # Realistic anisotropy profile from a Schwarzschild model.
    # The anisotropy varies smoothly between the following three regimes:
    # 1. beta = -1 for R < 1"
    # 2. beta = 0.3 for 1" < R < 30"
    # 3. beta = -0.2 for R > 30"
    #
    beta = np.empty_like(sigma_arcsec)
    beta[sigma_arcsec <= 1] = -1.0
    beta[(sigma_arcsec > 1) & (sigma_arcsec <= 30)] = 0.3
    beta[sigma_arcsec > 30] = -0.2

    # Compute V_RMS profiles and optimize M/L to best fit the data.
    # Assume self-consistency: same MGE for luminosity and potential.
    #
    pixSize = 0.1   # Spaxel size in arcsec
    sigmapsf = [0.1, 0.6]  # sigma of the PSF in arcsec from AO observations
    normpsf = [0.7, 0.3]
    mbh = 1.5e8     # Black hole mass in solar masses before multiplication by M/L
    distance = 20.  # Mpc

    rmsModel, ml, chi2 = jam_sph_rms(
        surf_pc, sigma_arcsec, surf_pc, sigma_arcsec, mbh, distance, rad,
        beta=beta, sigmapsf=sigmapsf, normpsf=normpsf, pixsize=pixSize,
        rms=sig, erms=erms, plot=True, tensor='los')

##############################################################################

if __name__ == '__main__':

    plt.clf()
    jam_sph_rms_example()
    plt.pause(1)



"""
#############################################################################

Copyright (C) 2014-2016, Michele Cappellari
E-mail: michele.cappellari_at_physics.ox.ac.uk

Updated versions of the software are available from my web page
http://purl.org/cappellari/software

If you have found this software useful for your research,
I would appreciate an acknowledgement to the use of the
"JAM modelling package of Cappellari (2008)"
or a reference to Note 11 of Cappellari et al. (2015).

This software is provided as is without any warranty whatsoever.
Permission to use, for non-commercial purposes is granted.
Permission to modify for personal or internal use is granted,
provided this copyright and disclaimer are included unchanged
at the beginning of the file. All other rights are reserved.
In particular, redistribution of the code is not allowed.

#############################################################################

MODIFICATION HISTORY:
    V1.0.0: Written and tested. Michele Cappellari, Oxford, 8 April 2014
    V1.0.1: All input distances are converted to pc. MC, Oxford, 15 May 2015
    V1.0.2: Updated documentation. MC, Oxford, 2 March 2016

"""

import numpy as np
from scipy.special import erf

def mge_radial_density(surf, sigma, qobs, inc, rad, distance=1e-5):
    """
    PURPOSE
    -------
    
    Calculates the spherically-averaged density of an axisymmetric MGE
    model, at a given set of radii, using the equation in Note 11 of
    `Cappellari et al. (2015, ApJL, 804, L21)
    <https://ui.adsabs.harvard.edu/abs/2015ApJ...804L..21C>`_

    INPUT PARAMETERS
    ----------------
    
    SURF: 
        Peak surface brightness of each Gaussian in Lsun/pc**2
    SIGMA: 
        Gaussian dispersion in arcsec
    QOBS: 
        Observed axial ratio of each Gaussian
    INC: 
        Inclination of the MGE model, with i=90 being edge-on
    RAD: 
        radius for which the density is requested (scalar or vector)

    OPTIONAL KEYWORDS
    -----------------
    
    DISTANCE: 
        Galaxy distance in Mpc (Default 10pc = 1e-5 Mpc)

    OUTPUTS:
        DENS: spherical density calculated at the input radii RAD

    """

    assert surf.size == sigma.size == qobs.size, "The MGE components do not match"

    rad = np.atleast_1d(rad)
    pc = distance*np.pi/0.648  # Constant factor to convert arcsec --> pc
    sigma = sigma*pc
    rad = rad*pc

    # total luminosity each Gaussian (eq.12 of Cappellari 2008, MNRAS, 390, 71)
    lum = 2.*np.pi*surf*qobs*sigma**2

    # Axisymmetric deprojection (eq.14 of Cappellari 2008, MNRAS, 390, 71)
    inc = np.radians(inc)
    qintr = qobs**2 - np.cos(inc)**2
    if np.any(qintr < 0):
        raise ValueError('Inclination too low q < 0')
    qintr = np.sqrt(qintr)/np.sin(inc)
    qintr = qintr.clip(0.001, 0.999) # Clip to avoid division by zero in dens
    
    # Average density rho(r) on spherical surfaces, without circularizing MGE.
    # Ths is the integral of a 3-dim axisymmetric Gaussian over the surface
    # of a sphere, divided by the surface of that sphere (4pi*r^2).
    # This is the formula in footnote 11 of Cappellari et al. (2015, ApJL)
    e = np.sqrt(1. - qintr**2)
    rad = rad[:, None]
    dens = np.sum(lum*np.exp(-0.5*(rad/sigma)**2)
                  * erf(e*rad/(np.sqrt(2)*qintr*sigma))
                  / (4*np.pi*e*rad*sigma**2), axis=1)

    return dens

#------------------------------------------------------------------------------ 




#!/usr/bin/env python

"""
V1.0.0: Michele Cappellari, Oxford, 17 April 2018

"""

import numpy as np
import matplotlib.pyplot as plt

from jampy.legacy.jam_axi_rms import jam_axi_rms
from jampy.legacy.jam_axi_vel import jam_axi_vel

def jam_axi_vel_example():
    """
    Usage example for jam_axi_vel().
    It takes about 2s on a 2.5 GHz computer

    """
    np.random.seed(123)
    xbin, ybin = np.random.uniform(low=[-55, -40], high=[55, 40], size=[1000, 2]).T

    inc = 60.                                                # assumed galaxy inclination
    r = np.sqrt(xbin**2 + (ybin/np.cos(np.radians(inc)))**2) # Radius in the plane of the disk
    a = 40                                                   # Scale length in arcsec
    vr = 2000*np.sqrt(r)/(r+a)                               # Assumed velocity profile
    vel = vr * np.sin(np.radians(inc))*xbin/r                # Projected velocity field
    sig = 8700/(r+a)                                         # Assumed velocity dispersion profile
    rms = np.sqrt(vel**2 + sig**2)                           # Vrms field in km/s

    surf = np.array([39483., 37158., 30646., 17759., 5955.1, 1203.5, 174.36, 21.105, 2.3599, 0.25493])
    sigma = np.array([0.153, 0.515, 1.58, 4.22, 10, 22.4, 48.8, 105, 227, 525])
    qObs = np.full_like(sigma, 0.57)

    distance = 16.5   # Assume Virgo distance in Mpc (Mei et al. 2007)
    mbh = 1e8 # Black hole mass in solar masses
    beta = np.full_like(surf, 0.2)

    surf_lum = surf # Assume self-consistency
    sigma_lum = sigma
    qobs_lum = qObs
    surf_pot = surf
    sigma_pot = sigma
    qobs_pot = qObs

    sigmapsf = [0.6, 1.2]
    normpsf = [0.7, 0.3]
    pixsize = 0.8
    goodbins = r > 10  # Arbitrarily exclude the center to illustrate how to use goodbins

    # First the M/L is determined by fitting the Vrms.
    # In general beta_z and the inclination will also be
    # fitted at this stage as described in Cappellari (2008)

    rmsModel, ml, chi2, flux = jam_axi_rms(
        surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
        inc, mbh, distance, xbin, ybin, plot=True, rms=rms, goodbins=goodbins,
        sigmapsf=sigmapsf, beta=beta, pixsize=pixsize, tensor='zz', normpsf=normpsf)
    plt.pause(1)

    # The velocity is fitted at the best fitting M/L, beta_z and
    # inclination determined at the previous stage

    surf_pot *= ml  # Scale the density by the best fitting M/L
    velModel, kappa, chi2, flux = jam_axi_vel(
        surf_lum, sigma_lum, qobs_lum, surf_pot, sigma_pot, qobs_pot,
        inc, mbh, distance, xbin, ybin, plot=True, vel=vel, goodbins=goodbins,
        sigmapsf=sigmapsf, beta=beta, pixsize=pixsize, component='z', normpsf=normpsf)
    plt.pause(1)

##############################################################################

if __name__ == '__main__':

    jam_axi_vel_example()



"""
#############################################################################

    Copyright (C) 2019, Michele Cappellari

    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from PyPi
    https://pypi.org/project/jampy


    If you have found this software useful for your research,
    I would appreciate an acknowledgement and a link to the website.

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

#############################################################################

V1.0.0: Ported to Python from Shampine's MATLAB version
        by Michele Cappellari, Oxford, 4 June 2019

"""
import numpy as np
import matplotlib.pyplot as plt

################################################################################

class quad2d:
    """
    quad2d Purpose
    ----------------

    quad2d numerically evaluate the integral of f(x, y) over a plane region.

    It implements the algorithm described in
    `Shampine (2008) <https://doi.org/10.1016/j.amc.2008.02.012>`_

    Calling Sequence
    ----------------

    .. code-block:: python

        res = quad2d(fun, [a, b], [c, d], epsrel=0, epsabs=1e-5,  args=())
        print(f"integral = {res.integ} +/- {res.errbnd}")
        print("function calls:", res.fcall)

    approximates the integral

    .. math::

        I = \int_a^b \int_c^d f(x,y) dy dx

    Input Parameters
    ----------------

    fun:
        Evaluates ``f(x,y)``.  It must accept arrays X and Y and return an
        array ``Z = f(X,Y)`` of corresponding values. These arrays are all 14x14.
        The region is a generalized rectangle which is described in Cartesian
        coordinates (x,y) by ``a <= x <= b`` -- A and B must be constants and
        ``c <= y <= d`` -- C and D can be constants or functions of x that
        describe the lower and upper boundaries. Functions must be vectorized.
        ``res.integ`` is an approximation to the integral with an estimated
        bound ``ERRBND`` on the error that satisfies  ``ERRBND <= 1e-5``.
    epsabs:
        Absolute error tolerance
    epsrel:
        Relative error tolerance

        quad2d attempts to satisfy ERRBND <= max(epsabs, epsrel*|Q|).
        This is absolute error control when |Q| is sufficiently small and
        relative error control when |Q| is larger. A default tolerance
        value is used when a tolerance is not specified. The default value
        of 'epsrel' is 1e-5. The default value of 'epsabs' is 0.
    singular: quad2d can be applied to f(x,y) that are singular on a
       boundary. With value True, this option causes quad2d to use
       transformations to weaken singularities for better performance.
    args:
        Parameters to pass to the user function ``fun``

    """
    def __init__(self, fun, xlim, ylim, epsrel=1e-5, epsabs=0,
                 plot=False, verbose=0, singular=False, args=()):

        a, b = xlim
        c, d = ylim
        if callable(c):
            self.phiBvar = c
        elif np.isscalar(c):
            self.phiBvar = lambda x: np.full_like(x, c)
        else:
            raise ValueError('C must be a constant or a function')

        if callable(d):
            self.phiTvar = d
        elif np.isscalar(d):
            self.phiTvar = lambda x: np.full_like(x, d)
        else:
            raise ValueError('D must be a constant or a function')

        if plot:
            plt.figure('quad2d')
            plt.clf()
            fig, self.axis = plt.subplots(2, gridspec_kw={'hspace': 0.02},
                                          sharex=True, sharey=True, num='quad2d')
            self.axis[0].set_ylabel('y')
            self.axis[1].set_xlabel('x')
            self.axis[1].set_ylabel('y')
            self.xall = self.yall = self.zall = []

        self.plot = plot
        self.EPS = 100.*np.finfo(float).eps
        self.singular = singular

        if singular:
            thetaL, thetaR = 0, np.pi
            phiB, phiT = 0, np.pi
        else:
            thetaL, thetaR = a, b
            phiB, phiT = 0, 1

        # Set defaults and then process optional values.
        self.FUN = fun
        self.fcall = 0
        self.AREA = (thetaR - thetaL)*(phiT - phiB)

        # Gauss-Kronrod (3,7) pair with degrees of precision 5 and 11.
        NODES = np.array([-0.9604912687080202, -0.7745966692414834, -0.4342437493468026,
                          0, 0.4342437493468026, 0.7745966692414834, 0.9604912687080202])
        self.NODES = np.hstack([NODES + 1, NODES + 3])/4
        WT3 = np.array([0, 5, 0, 8, 0, 5, 0])/9
        WT7 = np.array([0.1046562260264672, 0.2684880898683334, 0.4013974147759622,
                        0.4509165386584744, 0.4013974147759622, 0.2684880898683334,
                        0.1046562260264672])

        # Compute initial approximations on four subrectangles.  Initialize LIST
        # of information about subrectangles for which the approximations are
        # not sufficiently accurate.  NLIST is the number of subrectangles that
        # remain to be processed.  ERRBND is a bound on the error.
        Qsub, esub = self.tensor(a, b, thetaL, thetaR, phiB, phiT, WT3, WT7, args)
        self.integ = Qsub.sum()
        # Use an artificial value of TOL to force the program to refine.
        self.TOL = self.EPS*np.abs(self.integ)
        self.ERR_OK = 0
        self.ADJUST = 1
        self.LIST = np.zeros((200, 7))
        self.NLIST = -1
        self.Save2LIST(Qsub, esub, thetaL, thetaR, phiB, phiT)
        if self.NLIST < 0 or self.errbnd <= self.TOL:
            return

        while True:
            # Get entries from LIST corresponding to the biggest (adjusted) error.
            q, e, thetaL, thetaR, phiB, phiT = self.NextEntry()

            # Approximate integral over four subrectangles.
            Qsub, esub = self.tensor(a, b, thetaL, thetaR, phiB, phiT, WT3, WT7, args)

            # Saved in LIST is "e", a conservative estimate of the error in the
            # approximation "q" of the integral over a rectangle.  Newq = sum(Qsub)
            # is a much better approximation to the integral.  It is used here to
            # estimate the error in "q" and thereby determine that the estimator is
            # conservative by a factor of "ADJUST".  This factor is applied to the
            # estimate of the error in "Newq" to get a more realistic estimate.
            # This scheme loses the sign of the error, so a conservative local test
            # is used to decide convergence.
            Newq = Qsub.sum()
            self.ADJUST = min(1, abs(q - Newq)/e)
            self.integ += Newq - q
            TOL = max(epsabs, epsrel*abs(self.integ))/8
            TOL = max(TOL, self.EPS*abs(self.integ))
            self.Save2LIST(Qsub, esub, thetaL, thetaR, phiB, phiT)

            # Test for convergence and failures.
            if self.NLIST < 0 or self.errbnd <= TOL:
                break
            elif self.NLIST >= 2000:
                if self.errbnd > max(epsabs, max(self.EPS, epsrel)*abs(self.integ)):
                    raise ValueError(f'Maximum number of subintervals {self.NLIST}: Q does NOT pass error test.')
                else:
                    print(f'Maximum number of subintervals {self.NLIST}: Q appears to pass error test.')
                break

        if verbose:
            print(f"Integral: {self.integ} +/- {self.errbnd}")
            print(f"Function calls: {self.fcall} Function evaluations: {self.fcall*14**2}")

        if self.plot:
            self.axis[1].tricontourf(self.xall, self.yall, self.zall)
            self.axis[1].plot(self.xall, self.yall, 'w,')

################################################################################

    def tensor(self, a, b, thetaL, thetaR, phiB, phiT, WT3, WT7, args):
        """
        Compute the integral with respect to theta from thetaL to thetaR of the
        integral with respect to phi from phiB to phiT of F in four blocks.

        """
        n = WT7.size
        dtheta = thetaR - thetaL
        theta = thetaL + dtheta*self.NODES
        if self.singular:
            xx = 0.5*(b + a) + 0.5*(b - a)*np.cos(theta)
        else:
            xx = theta
        dphi = phiT - phiB
        phi = (phiB + dphi*self.NODES)[:, None]
        top = self.phiTvar(xx)
        bottom = self.phiBvar(xx)
        dydt = top - bottom
        if self.singular:
            t = 0.5 + 0.5*np.cos(phi)
        else:
            t = phi
        x = np.tile(xx, (2*n, 1))
        y = bottom + t*dydt
        z = self.FUN(x.ravel(), y.ravel(), *args)
        z = z.reshape(x.shape)
        self.fcall += 1
        if self.singular:
            z *= 0.25*(b - a)*np.sin(phi)*(dydt*np.sin(theta))
        else:
            z *= dydt

        # Split array into four n*n quadrants
        zz = z.reshape(2, n, 2, n).swapaxes(1, 2).reshape(4, n, n)

        # Four tensor products: Gauss 3 point formula
        esub = WT3 @ zz @ WT3
        esub = (esub/4)*(dtheta/2)*(dphi/2)

        # Four tensor products: Kronrod 7 point formula
        Qsub = WT7 @ zz @ WT7
        Qsub = (Qsub/4)*(dtheta/2)*(dphi/2)

        esub = np.abs(esub - Qsub)

        if self.plot:
            self.axis[0].plot(x.ravel(), y.ravel(), '+')
            self.xall = np.append(self.xall, x)
            self.yall = np.append(self.yall, y)
            self.zall = np.append(self.zall, z)

        return Qsub, esub

################################################################################

    def NextEntry(self):
        """
        Take entries from LIST corresponding to the biggest adjusted error

        """
        index = np.argmax(np.abs(self.LIST[:self.NLIST + 1, 6]))
        row = self.LIST[index, :6]
        self.LIST = np.delete(self.LIST, index, axis=0)
        self.NLIST -= 1

        return row

################################################################################

    def Save2LIST(self, Qsub, esub, thetaL, thetaR, phiB, phiT):
        """
        Save to the list information about subrectangles for which the
        integral is not sufficiently accurate. NLIST is the number of
        subrectangles to be processed.

        """
        dtheta = thetaR - thetaL
        dphi = phiT - phiB
        localtol = self.TOL*(dtheta/2)*(dphi/2)/self.AREA
        localtol = max(localtol, self.EPS*abs(Qsub.sum()))

        adjerr = self.ADJUST*esub
        if self.NLIST + 5 > len(self.LIST):
            self.LIST = np.vstack([self.LIST, np.zeros((100, 7))])

        if adjerr[0] > localtol:
            self.NLIST += 1
            self.LIST[self.NLIST] = [Qsub[0], esub[0], thetaL, thetaL + dtheta/2, phiB, phiB + dphi/2, adjerr[0]]
        else:
            self.ERR_OK += adjerr[0]

        if adjerr[1] > localtol:
            self.NLIST += 1
            self.LIST[self.NLIST] = [Qsub[1], esub[1], thetaL + dtheta/2, thetaR, phiB, phiB + dphi/2, adjerr[1]]
        else:
            self.ERR_OK += adjerr[1]

        if adjerr[2] > localtol:
            self.NLIST += 1
            self.LIST[self.NLIST] = [Qsub[2], esub[2], thetaL, thetaL + dtheta/2, phiB + dphi/2, phiT, adjerr[2]]
        else:
            self.ERR_OK += adjerr[2]

        if adjerr[3] > localtol:
            self.NLIST += 1
            self.LIST[self.NLIST] = [Qsub[3], esub[3], thetaL + dtheta/2, thetaR, phiB + dphi/2, phiT, adjerr[3]]
        else:
            self.ERR_OK += adjerr[3]

        self.errbnd = self.ERR_OK + self.LIST[:, 6].sum()

################################################################################
